from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.value import Value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.walkers import walkAllOriginSignals


def areInstanceOf(a, b, cls):
    return isinstance(a, cls) and isinstance(b, cls)

class ExprComparator():
    @staticmethod
    def isSimilar(exprA, exprB, diffInA):
        """
        @return:  tuple (match, exprDiffB)
        @attention: works only on simple expressions like constrain definition,
                    does not support multiple drivers for signals
        """
        if exprA  is diffInA:
            return (True, exprB)
        elif areInstanceOf(exprA, exprB, RtlSignalBase):
            try:
                originA = exprA.origin
                originB = exprB.origin
            except AttributeError:
                return (False, None)
            return ExprComparator.isSimilar(originA, originB, diffInA)
        elif areInstanceOf(exprA, exprB, Operator):
            if exprA.operator == exprB.operator:
                diff = None
                for opA, opB in zip(exprA.ops, exprB.ops):
                    m = ExprComparator.isSimilar(opA, opB, diffInA)
                    if not m[0]:
                        return (False, None)
                    if not m[1] is None:
                        assert diff is m[1] or diff is None
                        diff = m[1]
                return (True, diff)
        elif areInstanceOf(exprA, exprB, Value) and exprA._eq(exprB).val:
            return (True, None)
        return (False, None)
    
    @staticmethod    
    def findExprDiffInParam(exprA, exprB):
        params = list(walkAllOriginSignals(exprA))
        l = len(params)
        if l == 0:
            return 
        elif l == 1:
            m = ExprComparator.isSimilar(exprA, exprB, params[0])
            if m[0] and m[1] is not None:
                yield (params[0], m[1])
        else:
            raise NotImplementedError("Searching for multiple differences in expression")



from copy import copy
from operator import eq, ne, lt, gt, ge, le

from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.typeShortcuts import vecT
from hwt.hdlObjects.types.bitVal_bitOpsVldMask import vldMaskForXor, \
    vldMaskForAnd, vldMaskForOr
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BOOL, INT, BIT, SLICE
from hwt.hdlObjects.types.eventCapableVal import EventCapableVal
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.integerVal import IntegerVal
from hwt.hdlObjects.types.slice import Slice
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value, areValues
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.hdlObjects.types.bitValFunctions import boundryFromType, \
    bitsCmp__val, bitsCmp, bitsBitOp__val, bitsBitOp, bitsArithOp__val, bitsArithOp, \
    getMulResT
from hwt.bitmask import mask, selectBit, selectBitRange, bitSetTo, \
    setBitRange

class BitsVal(EventCapableVal):
    """
    @attention: operator on signals are using value operator functions as well 
    """
    def _convSign__val(self, signed):
        if self._dtype.signed == signed:
            return self
        selfSign = self._dtype.signed 
        v = self.clone()
        w = self._dtype.bit_length()
        msbVal = 1 << (w - 1)
        if selfSign and not signed:
            if v.val < 0:
                v.val += msbVal
        elif not selfSign and signed:
            if v.val >= msbVal:
                v.val -= (msbVal - 1)
            
        return v
    
    def _convSign(self, signed):
            if isinstance(self, Value):
                return self._convSign__val(signed)
            else:
                if self._dtype.signed == signed:
                    return self
                t = copy(self._dtype)
                t.signed = signed
                if signed is None:
                    cnv = AllOps.BitsAsVec
                elif signed:
                    cnv = AllOps.BitsAsSigned
                else:
                    cnv = AllOps.BitsAsUnsigned
                
                return Operator.withRes(cnv, [self], t)
        
    def _signed(self):
        return self._convSign(True)
    
    def _unsigned(self):
        return self._convSign(False)
    
    def _vec(self):
        return self._convSign(None)
                
    @classmethod
    def fromPy(cls, val, typeObj):
        assert isinstance(val, (int, bool)) or val is None, val
        vld = 0 if val is None else mask(typeObj.bit_length())
        if not vld:
            val = 0
        return cls(int(val), typeObj, vld)
    
    def _concat__val(self, other):
        w = self._dtype.bit_length()
        other_w = other._dtype.bit_length()
        resWidth = w + other_w
        resT = vecT(resWidth)
        
        v = self.clone()
        v.val = (v.val << other_w) | other.val
        v.vldMask = (v.vldMask << other_w) | other.vldMask
        v.updateTime = max(self.updateTime, other.updateTime)
        v._dtype = resT
        
        return v 
    
    def _concat(self, other):
        w = self._dtype.bit_length()
        other_w = other._dtype.bit_length()
        resWidth = w + other_w
        resT = vecT(resWidth)
        
        if areValues(self, other):
            return self._concat__val(other)
        else:
            w = self._dtype.bit_length()
            other_w = other._dtype.bit_length()
            resWidth = w + other_w
            resT = vecT(resWidth)
            # is instance of signal
            if isinstance(other, InterfaceBase):
                other = other._sig
            if isinstance(other._dtype, Bits):
                if other._dtype.signed != None:
                    other = other._vec()
            elif other._dtype == BOOL:
                other = other._convert(BIT)
            else:
                raise TypeError(other._dtype)
            
            if self._dtype.signed != None:
                    self = self._vec()
            
            return Operator.withRes(AllOps.CONCAT, [self, other], resT)\
                           ._convert(vecT(resWidth, signed=self._dtype.signed))
    
    def _getitem__val(self, key):
        updateTime = max(self.updateTime, key.updateTime)
        keyVld = key._isFullVld() 
        val = 0
        vld = 0
        
        if isinstance(key._dtype, Integer):
            if keyVld:
                val = selectBit(self.val, key.val)
                vld = selectBit(self.vldMask, key.val)
            return BitsVal(val, BIT, vld, updateTime=updateTime)
        elif key._dtype == SLICE:
            if keyVld:
                firstBitNo = key.val[1].val
                size = key._size()
                val = selectBitRange(self.val, firstBitNo, size)
                vld = selectBitRange(self.vldMask, firstBitNo, size)
            retT = vecT(size, signed=self._dtype.signed)
            return BitsVal(val, retT, vld, updateTime=updateTime)
        else:
            raise TypeError(key)
    
    def __getitem__(self, key):
        iamVal = isinstance(self, Value)
        st = self._dtype
        l = st.bit_length()
        if l == 1:
            assert st.forceVector  # assert not indexing on single bit
            
        isSlice = isinstance(key, slice)
        isSLICE = isinstance(key, Slice.getValueCls())
        if areValues(self, key):
            return self._getitem__val(key)
        elif isSlice or isSLICE:
            if isSlice:
                if key.step is not None:
                    raise NotImplementedError()
                start = key.start
                stop = key.stop
                
                if key.start is None:
                    start = boundryFromType(self, 0) + 1
                else:
                    start = toHVal(key.start)
                
                if key.stop is None:
                    stop = boundryFromType(self, 1)
                else:
                    stop = toHVal(key.stop)
            else:
                start = key.val[0]
                stop = key.val[1]

            isVal = iamVal and isinstance(start, Value) and isinstance(stop, Value) 
            if isVal:
                raise NotImplementedError("[TODO] bit select on value")
            else:
                key = (start - INT.fromPy(1))._downto(stop)
                # [TODO] type can be wrong, but we need to get rid off widthConstr and use only width
                _resWidth = (start - 1 - stop)._downto(0)
                resT = Bits(widthConstr=_resWidth, forceVector=True, signed=st.signed)
                
        elif isinstance(key, (int, IntegerVal)):
            key = toHVal(key)
            resT = BIT
            
        elif isinstance(key, RtlSignalBase):
            t = key._dtype
            if isinstance(t, Integer):
                resT = BIT
            elif isinstance(t, Slice):
                resT = Bits(widthConstr=key, forceVector=st.forceVector, signed=st.signed)
            elif isinstance(t, Bits):
                resT = BIT
                key = key._convert(INT)
            else:
                raise TypeError("Index operation not implemented for index of type %s" % (repr(t)))
       
        else:
            raise TypeError("Index operation not implemented for index %s" % (repr(key)))
            
        # [TODO] boundary check
        return Operator.withRes(AllOps.INDEX, [self, key], resT)

    def _setitem__val(self, index, value):
        if index._isFullVld():
            if isinstance(index._dtype, Integer): 
                self.val = bitSetTo(self.val, index.val, value.val)
                self.vldMask = bitSetTo(self.vldMask, index.val, value.vldMask)
            elif index._dtype == SLICE:
                size = index._size()
                noOfFirstBit = index.val[1].val
                self.val = setBitRange(self.val, noOfFirstBit, size, value.val)
                self.vldMask = setBitRange(self.vldMask, noOfFirstBit, size, value.vldMask)
            else:
                raise TypeError("Not implemented for index %s" % repr(index))
            self.updateTime = max(index.updateTime, value.updateTime)
        else:
            self.vldMask = 0

    def __setitem__(self, index, value):
        assert isinstance(self, Value)
        return self._setitem__val(index, value)

    def _invert__val(self):
        v = self.clone()
        v.val = ~v.val
        w = v._dtype.bit_length()
        v.val &= mask(w)
        return v
    
    def __invert__(self):
        if isinstance(self, Value):
            return self._invert__val()
        else:
            try:
                # double negation
                d = self.singleDriver()
                if isinstance(d, Operator) and d.operator == AllOps.NOT:
                    return d.ops[0]
            except MultipleDriversExc:
                pass
            return Operator.withRes(AllOps.NOT, [self], self._dtype)
    
    # comparisons         
    def _eq__val(self, other):
        return bitsCmp__val(self, other, AllOps.EQ, eq)
    def _eq(self, other):
        return bitsCmp(self, other, AllOps.EQ, eq)
    
    
    def _ne__val(self, other):
        return bitsCmp__val(self, other, AllOps.NEQ, ne)
    def __ne__(self, other):
        return bitsCmp(self, other, AllOps.NEQ)
    

    def _lt__val(self, other):
        return bitsCmp__val(self, other, AllOps.LOWERTHAN, lt)
    def __lt__(self, other):
        return bitsCmp(self, other, AllOps.LOWERTHAN)
    
    
    def _gt__val(self, other):
        return bitsCmp__val(self, other, AllOps.GREATERTHAN, gt)
    def __gt__(self, other):
        return bitsCmp(self, other, AllOps.GREATERTHAN)

    
    def _ge__val(self, other):
        return bitsCmp__val(self, other, AllOps.GE, ge)
    def __ge__(self, other):
        return bitsCmp(self, other, AllOps.GE)
   
   
    def _le__val(self, other):
        return bitsCmp__val(self, other, AllOps.LE, le)
    def __le__(self, other):
        return bitsCmp(self, other, AllOps.LE)
    
    
    def _xor__val(self, other):
        return bitsBitOp__val(self, other, AllOps.XOR, vldMaskForXor)
    def __xor__(self, other):
        return bitsBitOp(self, other, AllOps.XOR, vldMaskForXor)
    
    
    def _and__val(self, other):
        return bitsBitOp__val(self, other, AllOps.AND_LOG, vldMaskForAnd)
    def __and__(self, other):
        return bitsBitOp(self, other, AllOps.AND_LOG, vldMaskForAnd)
    
    
    def _or__val(self, other):
        return bitsBitOp__val(self, other, AllOps.OR_LOG, vldMaskForOr)
    def __or__(self, other):
        return bitsBitOp(self, other, AllOps.OR_LOG, vldMaskForOr)
       

    def _sub__val(self, other):
        return bitsArithOp__val(self, other, AllOps.SUB)
    def __sub__(self, other):
        return bitsArithOp(self, other, AllOps.SUB)
    
   
    def _add__val(self, other):
        return bitsArithOp__val(self, other, AllOps.ADD)
    def __add__(self, other):
        return bitsArithOp(self, other, AllOps.ADD)
    
    def _mul__val(self, other):
        resT = getMulResT(self._dtype, other._dtype)
        val = self.val * other.val
        result = resT.fromPy(val)
        
        raise NotImplementedError() 
        # # [TODO] value check range
        # if isinstance(other._dtype, Integer):
        #    if other.vldMask:
        #        v.vldMask = self.vldMask
        #    else:
        #        v.vldMask = 0
        #
        # elif isinstance(other._dtype, Bits):
        #    v.vldMask = self.vldMask & other.vldMask
        # else:
        #    raise TypeError("Incompatible type for multiplication: %s" % (repr(other._dtype)))
        # # [TODO] correct overflow detection for signed values
        # w = v._dtype.bit_length()
        # v.val &= Bitmask.mask(2*w)
        # v.updateTime = max(self.updateTime, other.updateTime)
        # return v
    def __mul__(self, other):        
        other = toHVal(other)
        assert isinstance(other._dtype, (Integer, Bits))
        
        
        if areValues(self, other):
            return self._mul__val(other)
        else:
            resT = getMulResT(self._dtype, other._dtype)
            if self._dtype.signed is None:
                self = self._unsigned()
            if isinstance(other._dtype, Bits) and other._dtype.signed is None:
                other = other._unsigned() 
            elif isinstance(other._dtype, Integer):
                pass
            else:
                raise TypeError("%s %s %s" % (repr(self), repr(AllOps.MUL) , repr(other)))
            
            subResT = vecT(resT.bit_length(), self._dtype.signed)
            o = Operator.withRes(AllOps.MUL, [self, other], subResT)
            return o._convert(resT)
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class Architecture(object):
    """
    Hdl container of internal structure of unit
    """
    def __init__(self, entity):
        self.entity = entity
        if entity: 
            self.entityName = entity.name 
        else:
            self.entityName = None
        self.name = "rtl"
        self.variables = []
        self.processes = []
        self.components = []
        self.componentInstances = []
    
    def getEntityName(self):
        if self.entity:
            return self.entity.name
        else:
            return self.entityName    
        
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer
        return VhdlSerializer.formater(VhdlSerializer.Architecture(self, VhdlSerializer.getBaseNameScope()))





class UniqList(list):
    """
    List of unique items
    """
    def __init__(self, *args, **kwargs):
        super(UniqList, self).__init__(*args, **kwargs)
        self.__s = set()
    
    def append(self, item):
        if item in self.__s:
            return False
        else:
            self.__s.add(item)
            list.append(self, item)
            return True



    
class Value():
    """
    Wrap around hdl value with overloaded operators
    http://www.rafekettler.com/magicmethods.html
    
    operators are overloaded in every type separately
    """
    __slots__ = ["val", "_dtype", "vldMask", "updateTime"]
    
    def __init__(self, val, _type, vldMask, updateTime=-1):
        self.val = val
        self._dtype = _type
        self.vldMask = vldMask
        self.updateTime = updateTime
    
    def _isFullVld(self):
        return self.vldMask == self._dtype.all_mask()

    def _convert(self, toT):
        return self._dtype.convert(self, toT)

    def staticEval(self):
        return self.clone()
    
    def simEval(self, simulator):
        return self.clone()
    
    def clone(self):
        return self.__class__(self.val, self._dtype, self.vldMask, self.updateTime)
    
    def __hash__(self):
        return hash((self._dtype, self.val, self.vldMask, self.updateTime))
    
    def __repr__(self):
        return "<Value {0:s}, mask {1:b}, time {2:.2f}>".format(
                    str(self.val), self.vldMask, self.updateTime)    

    @classmethod
    def fromPy(cls, val, typeObj):
        raise NotImplementedError("fromPy fn is not implemented for %s" % (str(cls)))
   
    def __nonzero__(self):
        raise NotImplementedError()
    def __bool__(self):
        raise NotImplementedError()
    def __pos__(self):
        raise NotImplementedError()
    def __neg__(self):
        raise NotImplementedError()    
    def __abs__(self):
        raise NotImplementedError()
    def __invert__(self):
        raise NotImplementedError()
    def __round__(self, n):
        raise NotImplementedError()
    def __floor__(self):
        raise NotImplementedError()
    def __ceil__(self):
        raise NotImplementedError()
    def __trunc__(self):
        raise NotImplementedError()
    def __add__(self, other):
        raise NotImplementedError()
    def __sub__(self, other):
        raise NotImplementedError()
    def __mul__(self, other):
        raise NotImplementedError()
    def __floordiv__(self, other):
        raise NotImplementedError()
    def __div__(self, other):
        raise NotImplementedError()
    def __truediv__(self, other):
        raise NotImplementedError()
    def __mod__(self, other):
        raise NotImplementedError()
    def __divmod__(self, other):
        raise NotImplementedError()
    def __pow__(self, other):
        raise NotImplementedError()
    def __lshift__(self, other):
        raise NotImplementedError()
    def __rshift__(self, other):
        raise NotImplementedError()
    def __and__(self, other):
        raise NotImplementedError()
    def __or__(self, other):
        raise NotImplementedError()
    def __xor__(self, other):
        raise NotImplementedError()
    def __eq__(self, other):
        if areValues(self, other):
            return self._dtype == other._dtype and  bool(self._eq(other))
        else:
            super().__eq__(other)
    
    def _eq(self, other):
        raise NotImplementedError()
    def __ne__(self, other):
        eq = self._eq(other)
        eq.val = not eq.val
        return eq
    def _concat(self, other):
        raise NotImplementedError()
    def __lt__(self, other):
        raise NotImplementedError()
    def __le__(self, other):
        raise NotImplementedError()
    def __gt__(self, other):
        raise NotImplementedError()
    def __ge__(self, other):
        raise NotImplementedError()
        
    def _hasEvent(self, now):
        raise NotImplementedError()
    
    def _onRisingEdge(self, now):
        raise NotImplementedError()
    
def areValues(*items):
    res = True
    for i in items:
        res = res and isinstance(i, Value)
    return res



from hwt.hdlObjects.constants import Time
from hwt.simulator.agentBase import AgentBase
from hwt.simulator.shortcuts import onRisingEdge


DEFAULT_CLOCK = 10 * Time.ns


class SignalAgent(AgentBase):
    """
    Agent for signal interface, it can use clock and reset interface for synchronization
    or can be synchronized by delay

    @attention: clock synchronization has higher priority
    """
    def __init__(self, intf, clk=None, rstn=None, delay=DEFAULT_CLOCK):
        self.delay = delay
        self.initDelay = 0
        self.clk = clk
        self.rstn = rstn
        self.intf = intf
        self.data = []

        self.initPending = True

        if clk is not None:
            self.monitor = onRisingEdge(self.clk, self.monitor)
            self.driver = onRisingEdge(self.clk, self.driver)

    def doRead(self, s):
        return s.read(self.intf)

    def doWrite(self, s, data):
        s.w(data, self.intf)

    def driver(self, s):
        if self.initPending and self.initDelay:
            yield s.wait(self.initDelay)
            self.initPending = False

        if self.clk is None:
            # if clock is specified this function is periodicaly called every
            # clk tick
            while True:
                if self.data:
                    self.doWrite(s, self.data.pop(0))
                yield s.wait(self.delay)
        else:
            # if clock is specified this function is periodicaly called every
            # clk tick
            if self.data:
                self.doWrite(s, self.data.pop(0))

    def monitor(self, s):
        if self.clk is None:
            if self.initPending and self.initDelay:
                yield s.wait(self.initDelay)
                self.initPending = False
            # if there is no clk, we have to manage periodic call by our selfs
            while True:
                yield s.updateComplete
                d = self.doRead(s)
                self.data.append(d)
                yield s.wait(self.delay)
        else:
            # if clock is specified this function is periodicaly called every
            # clk tick
            yield s.updateComplete
            d = self.doRead(s)
            self.data.append(d)



from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value, areValues
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.typeShortcuts import vecT
from hwt.bitmask import mask

BoolVal = BOOL.getValueCls()

def bitsCmp__val(self, other, op, evalFn):
    w = self._dtype.bit_length()
    assert w == other._dtype.bit_length(), "%d, %d" % (w, other._dtype.bit_length())
    
    vld = self.vldMask & other.vldMask
    _vld = vld == mask(w)
    res = evalFn(self.val, other.val) and _vld
    updateTime = max(self.updateTime, other.updateTime)
    
    return BoolVal(res, BOOL, int(_vld), updateTime)

def bitsCmp(self, other, op, evalFn=None):
    """
    @attention: If other is Bool signal convert this to boolean (not ideal, due VHDL event operator)
    """
    other = toHVal(other)
    
    iamVal = isinstance(self, Value)
    otherIsVal = isinstance(other, Value) 

    if evalFn is None:
        evalFn = op._evalFn
    
    if iamVal and otherIsVal:
        return bitsCmp__val(self, other, op, evalFn)
    else:
        if other._dtype == BOOL:
            self = self._convert(BOOL)
        elif self._dtype == other._dtype:
            pass
        elif isinstance(other._dtype, Integer):
            other = other._convert(self._dtype) 
        else:
            raise TypeError("Values of types (%s, %s) are not comparable" % (repr(self._dtype), repr(other._dtype)))
        
        return Operator.withRes(op, [self, other], BOOL) 

def bitsBitOp__val(self, other, op, getVldFn):
    w = self._dtype.bit_length()
    assert w == other._dtype.bit_length()
    
    vld = getVldFn(self, other)
    res = op._evalFn(self.val, other.val) & vld
    updateTime = max(self.updateTime, other.updateTime)
    
    return self.__class__(res, self._dtype, vld, updateTime)

def bitsBitOp(self, other, op, getVldFn):
    """
    @attention: If other is Bool signal, convert this to boolean (not ideal, due VHDL event operator)
    """
    other = toHVal(other)
    
    iamVal = isinstance(self, Value)
    otherIsVal = isinstance(other, Value) 
    
    if iamVal and otherIsVal:
        return bitsBitOp__val(self, other, op, getVldFn)
    else:
        if other._dtype == BOOL:
            self = self._convert(BOOL)
            return op._evalFn(self,  other)
        elif self._dtype == other._dtype:
            pass
        else:
            raise TypeError("Types are not comparable (%s, %s)" % 
                                       (repr(self._dtype), repr(other._dtype)))
        
        return Operator.withRes(op, [self, other], self._dtype) 

def bitsArithOp__val(self, other, op):
    v = self.clone()
    self_vld = self._isFullVld()
    other_vld = other._isFullVld()
     
    v.val = op._evalFn(self.val, other.val)

    # [TODO] correct overflow detection for signed values
    w = v._dtype.bit_length()
    v.val &= mask(w)
    
    # [TODO] value check range
    if self_vld and other_vld:
        v.vldMask = mask(w)
    else:
        v.vldMask = 0
        
    v.updateTime = max(self.updateTime, other.updateTime)
    return v

def bitsArithOp(self, other, op):
    other = toHVal(other)
    assert isinstance(other._dtype, (Integer, Bits))
    if areValues(self, other):
        return bitsArithOp__val(self, other, op)
    else:
        resT = self._dtype
        if self._dtype.signed is None:
            self = self._unsigned()
        if isinstance(other._dtype, Bits):
            other = other._convSign(self._dtype.signed)    
        elif isinstance(other._dtype, Integer):
            pass
        else:
            raise TypeError("%s %s %s" % (repr(self), repr(op) , repr(other)))
        
        o = Operator.withRes(op, [self, other], self._dtype)
        return o._convert(resT)

def boundryFromType(sigOrVal, boundaryIndex):
    c = sigOrVal._dtype.constrain
    if isinstance(c, Value):  # slice
        return c.val[boundaryIndex]
    else:  # downto / to
        return c.singleDriver().ops[boundaryIndex]

def getMulResT(firstT, secondT):
    if isinstance(secondT, Integer):
        return firstT # [maybe wrong]
    
    width = firstT.bit_length() + secondT.bit_length()
    return vecT(width, firstT.signed)
    



from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.types.bits import Bits

def pushBit(v, b):
    return (v << 1) | b

def convertString(self, sigOrVal, toType):
    if isinstance(toType, Bits):
        if isinstance(sigOrVal, Value):
            v = sigOrVal.clone()
            _v = v.val
            v.val = 0
            v.vldMask = 0
            v._dtype = toType
            for ch in reversed(_v): 
                if ch == '1':
                    v.val = pushBit(v.val, 1)
                    v.vldMask = pushBit(v.vldMask, 1)
                elif ch == '0':
                    v.val = pushBit(v.val, 0)
                    v.vldMask = pushBit(v.vldMask, 1)
                elif ch == 'x':
                    v.val = pushBit(v.val, 0)
                    v.vldMask = pushBit(v.vldMask, 0)
                else:
                    raise NotImplementedError("found %s in bitstring literal" % (ch))
            return v
        
    return HdlType.defaultConvert(self, sigOrVal, toType)  


from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.statements import IfContainer, SwitchContainer
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.value import Value
from hwt.synthesizer.assigRendererContainers import IfTreeNode
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase 
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.synthesizer.termUsageResolver import extractCondTermOrderNonResolved, extractCondTermOrder
from hwt.pyUtils.arrayQuery import where

# (max count of elsifs with eq on same variable)
SWITCH_THRESHOLD = 2  

def __renderStatements(statements, resultContainer):
    for st in statements:
        if isinstance(st, IfTreeNode):
            for o in _renderIfTree(st): 
                resultContainer.append(o)
        else:
            resultContainer.append(st)

def _renderIfTree(node):
    """
    Render tree of IfTreeNode objects to IfContainers, SwitchContainers etc.
    
    if number of elsifs is higher than SWITCH_THRESHOLD and all conditions are if format signal == value
    render this as switch statement
    """
    assert isinstance(node, IfTreeNode), node  
    
    ifTrue = []
    __renderStatements(node.pos, ifTrue)
    
    elIfs = []
    ifFalse = []
    
    # extract elsifs
    elIfN = node
    while True:
        if len(elIfN.neg) == 1 and isinstance(elIfN.neg[0], IfTreeNode):
            # if only next if is in negative statements
            # this item is converted as elsif in this IfContainer
            elIfN = elIfN.neg[0]
            _ifTrue = []
            __renderStatements(elIfN.pos, _ifTrue)
            elIfs.append(([elIfN.cond, ], _ifTrue))
        else:
            # render standard else
            __renderStatements(elIfN.neg, ifFalse)
            break
    
    # if threshold
    if len(elIfs) >= SWITCH_THRESHOLD:
        cases = []
        switchOn = None
        try:
            # detect if condition is in format signal == value 
            op = node.cond.singleDriver()
            if isinstance(op, Operator) and op.operator == AllOps.EQ:
                op0 = op.ops[0]
                op1 = op.ops[1]
                if isinstance(op0, RtlSignalBase) and isinstance(op1, Value):
                    switchOn = op0
                    cases.append((op1, ifTrue))
        except MultipleDriversExc:
            pass
        
        if switchOn is not None:
            canBeConvertedToSwitch = True
            for elIf in elIfs:
                try:
                    op = elIf[0][0].singleDriver()
                    if isinstance(op, Operator) and op.operator == AllOps.EQ:
                        op1 = op.ops[1]
                        if op.ops[0] is switchOn and isinstance(op1, Value):
                            cases.append((op1, elIf[1]))
                            continue
                    canBeConvertedToSwitch = False
                    break
                except MultipleDriversExc:
                    canBeConvertedToSwitch = False
            
            # if only last can not be part of the swicht case it can be default 
            if not canBeConvertedToSwitch and len(elIfs) == len(cases):
                default = elIfs[-1]
                ifFalse = [ IfContainer(default[0], default[1], ifFalse)]
                canBeConvertedToSwitch = True
                    
            if canBeConvertedToSwitch:
                t = switchOn._dtype
                # if nothing else and we have enum ad we used all the values
                if not ifFalse and isinstance(t, Enum) and len(t._allValues) == len(cases):
                    # convert last to default, because hdl languages usually need this
                    ifFalse = cases[-1][1]
                    cases = cases[:-1]
                    
                cases.append((None, ifFalse))
            
                yield SwitchContainer(switchOn, cases)
                raise StopIteration()
    
    yield IfContainer([node.cond], ifTrue, ifFalse, elIfs=elIfs)

def renderIfTree_afterCondSatisfied(assignments, globalCondOrder):
    # there can be only one main condition ant this means 
    # only the most impact cond is taken
    # else no condition is taken and assignment is just statement at this level
    # in this step we are consuming assignment as statement
    top, topConds = extractCondTermOrderNonResolved(assignments, globalCondOrder)

    if len(topConds) == 0:
        return top 
    else:
        topCond = topConds[0]
        
        # remove resolved statements from unresolved container
        for a in top:
            assignments.remove(a)
            
        # create IfTreeNode for topCond
        ifs = []
        topIf, notDependent = splitIfTreeOnCond(assignments, topCond, topConds[1:])
        ifs.append(topIf)
        if notDependent:
            _ifs = renderIfTree_afterCondSatisfied(notDependent, globalCondOrder)
            ifs.extend(_ifs)
            
        return ifs
    
def splitIfTreeOnCond(assignments, topCond, globalCondOrder):
    # in this step we are consuming unresolvedConds and building IfTreeNodes

    topPos = []
    topNeg = []
    notDependent = []
    for a in assignments:
        dependentOnTopCond = list(where(a._unresolvedConds, lambda cond: cond[0] is topCond))
        l = len(dependentOnTopCond)
        
        if l == 0:
            notDependent.append(a)
        elif l == 1:
            _c = dependentOnTopCond[0]
            a._unresolvedConds.remove(_c)
            
            if _c[1]:
                topNeg.append(a)
            else:
                topPos.append(a)
        else:
            raise NotImplementedError(
                ("There are multiple different assignments to same object(%r) with same condition (%r)" % (a.dst, topCond),
                dependentOnTopCond))     
    # if not (len(assignments) == (len(topNeg) + len(topPos))):
    #    # it seems that there is some statement which is nod depended on topCond, but it should be 
    #    # filtered earlier
    #    raise AssertionError(("got assignments %s and topCond %s \n for neg resolved %s,\n" + 
    #                         " for pos resolved %s\n" + 
    #                         "something lost or duplicited in statement renderer") 
    #                         % (str(assignments), str(topCond), str(topNeg), str(topPos)))
    if len(topPos) + len(topNeg) == 0:
        raise AssertionError("Something should be dependent")
    
    top = IfTreeNode(topCond)
    top.pos = renderIfTree_afterCondSatisfied(topPos, globalCondOrder)
    top.neg = renderIfTree_afterCondSatisfied(topNeg, globalCondOrder)
    return top, notDependent


def renderIfTree(assignments):
    """
    Walk assignments and resolve if tree from conditions
    """
    condOrder = list(extractCondTermOrder(assignments))
    
    if condOrder:
        # split assignments on most important condition
        topCond = condOrder[0]
        top, notDependent = splitIfTreeOnCond(assignments, topCond, condOrder)
        
        yield from _renderIfTree(top)
        if notDependent:
            yield from renderIfTree(notDependent)
    else:
        # none of assignments has condition no If or switch is needed
        yield from assignments



from hwt.hdlObjects.types.hdlType import HdlType 
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import INT
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.bitmask import mask

def convertBits__val(self, sigOrVal, toType):
    if isinstance(toType, Boolean):
        return sigOrVal._eq(self.getValueCls().fromPy(1, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT:
        if self.signed:
            raise NotImplementedError()
        else:
            fullMask = mask(self.bit_length())
            return INT.getValueCls()(sigOrVal.val, INT, sigOrVal.vldMask == fullMask, sigOrVal.updateTime)
    return HdlType.defaultConvert(self, sigOrVal, toType)


def convertBits(self, sigOrVal, toType):
    isVal = isinstance(sigOrVal, Value)
    
    if isinstance(toType, Boolean):
        if isVal:
            return sigOrVal._eq(self.getValueCls().fromPy(1, self))
        elif self.bit_length() == 1:
            v = 0 if sigOrVal.negated else 1
            return sigOrVal._eq(self.getValueCls().fromPy(v, self))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            return sigOrVal._convSign(toType.signed)
    elif toType == INT:
        if isVal:
            if self.signed:
                raise NotImplementedError()
            else:
                fullMask = mask(self.bit_length())
                return INT.getValueCls()(sigOrVal.val, INT, sigOrVal.vldMask == fullMask, sigOrVal.updateTime)
        else:
            return Operator.withRes(AllOps.BitsToInt, [sigOrVal], toType)

    return HdlType.defaultConvert(self, sigOrVal, toType)



from hwt.hdlObjects.constants import Time
from hwt.simulator.agentBase import AgentBase
from hwt.simulator.shortcuts import pullUpAfter, pullDownAfter


class PullUpAgent(AgentBase):
    def __init__(self, intf, intDelay=6 * Time.ns):
        self.intDelay = intDelay
        self.data = []
        self.driver = pullUpAfter(intf, intDelay=intDelay)
        
class PullDownAgent(AgentBase):
    def __init__(self, intf, intDelay=6 * Time.ns):
        self.intDelay = intDelay
        self.data = []
        self.driver = pullDownAfter(intf, intDelay=intDelay)
        



from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdlObjects.operatorDefs import AllOps, OpDefinition


# keep in mind that there is no such a thing in vhdl itself
opPrecedence = {AllOps.NOT : 2,
                AllOps.EVENT: 1,
                AllOps.RISING_EDGE: 1,
                AllOps.DIV: 3,
                AllOps.ADD : 3,
                AllOps.SUB: 3,
                AllOps.MUL: 3,
                AllOps.MUL: 3,
                AllOps.XOR: 2,
                AllOps.EQ: 2,
                AllOps.NEQ: 2,
                AllOps.AND_LOG: 2,
                AllOps.OR_LOG: 2,
                AllOps.DOWNTO: 2,
                AllOps.GREATERTHAN: 2,
                AllOps.LOWERTHAN: 2,
                AllOps.CONCAT: 2,
                AllOps.INDEX: 1,
                AllOps.TERNARY: 1,
                AllOps.CALL: 1,
                }

def isResultOfTypeConversion(sig):
    try:
        d = sig.drivers[0]
    except IndexError:
        return False
    
    if sig.hidden:
        return True
    return False

class VhdlSerializer_ops():

    @classmethod
    def BitToBool(cls, cast):
        v = 0 if cast.sig.negated else 1
        return cls.asHdl(cast.sig) + "=='%d'" % v

    
    @classmethod
    def Operator(cls, op, createTmpVarFn):
        def p(operand):
            s = cls.asHdl(operand, createTmpVarFn)
            if isinstance(operand, RtlSignalBase):
                try:
                    o = operand.singleDriver()
                    if opPrecedence[o.operator] <= opPrecedence[op.operator]:
                        return " (%s) " % s
                except Exception:
                    pass
            return " %s " % s
        
        ops = op.ops
        o = op.operator
        def _bin(name):
            return (" " + name + " ").join(map(lambda x: x.strip(), map(p, ops)))
        
        if o == AllOps.AND_LOG:
            return _bin('AND')
        elif o == AllOps.OR_LOG:
            return _bin('OR')
        elif o == AllOps.XOR:
            return _bin('XOR')
        elif o == AllOps.NOT:
            assert len(ops) == 1
            return "NOT " + p(ops[0])
        elif o == AllOps.CALL:
            return "%s(%s)" % (cls.FunctionContainer(ops[0]), ", ".join(map(p, ops[1:])))
        elif o == AllOps.CONCAT:
            return _bin('&')
        elif o == AllOps.DIV:
            return _bin('/')
        elif o == AllOps.DOWNTO:
            return _bin('DOWNTO')
        elif o == AllOps.EQ:
            return _bin('=')
        elif o == AllOps.EVENT:
            assert len(ops) == 1
            return p(ops[0]) + "'EVENT"
        elif o == AllOps.GREATERTHAN:
            return _bin('>')
        elif o == AllOps.GE:
            return _bin('>=')
        elif o == AllOps.LE:
            return _bin('<=')
        elif o == AllOps.INDEX:
            assert len(ops) == 2
            o1 = ops[0]
            if isResultOfTypeConversion(o1):
                o1 = createTmpVarFn("tmpTypeConv", o1._dtype)
                o1.defaultVal = ops[0]
            
            return "%s(%s)" % (cls.asHdl(o1, createTmpVarFn).strip(), p(ops[1]))
        elif o == AllOps.LOWERTHAN:
            return _bin('<')
        elif o == AllOps.SUB:
            return _bin('-')
        elif o == AllOps.MUL:
            return _bin('*')
        elif o == AllOps.NEQ:
            return _bin('/=')
        elif o == AllOps.ADD:
            return _bin('+')
        elif o == AllOps.TERNARY:
            return p(ops[1]) + " WHEN " + cls.condAsHdl([ops[0]], True, createTmpVarFn) + " ELSE " + p(ops[2])
        elif o == AllOps.RISING_EDGE:
            assert len(ops) == 1
            return "RISING_EDGE(" + p(ops[0]) + ")"
        elif o == AllOps.FALLIGN_EDGE:
            assert len(ops) == 1
            return "FALLING_EDGE(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsSigned:
            assert len(ops) == 1
            return  "SIGNED(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsUnsigned:
            assert len(ops) == 1
            return  "UNSIGNED(" + p(ops[0]) + ")"
        elif o == AllOps.BitsAsVec:
            assert len(ops) == 1
            return  "STD_LOGIC_VECTOR(" + p(ops[0]) + ")"
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = cls.asHdl(ops[0], createTmpVarFn)
            if ops[0]._dtype.signed is None:
                op = "UNSIGNED(%s)" % op
            return "TO_INTEGER(%s)" % op
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            op_str = cls.asHdl(ops[0], createTmpVarFn)
            w = resT.bit_length()
            
            if resT.signed is None:
                return "STD_LOGIC_VECTOR(TO_UNSIGNED(" + op_str + ", %d))" % (w)
            elif resT.signed:
                return "TO_UNSIGNED(" + op_str + ", %d)" % (w)
            else:
                return "TO_UNSIGNED(" + op_str + ", %d)" % (w)
            
        elif o == AllOps.POW:
            assert len(ops) == 2
            return _bin('**')
        else:
            raise NotImplementedError("Do not know how to convert %s to vhdl" % (o))




from hwt.hdlObjects.constants import DIRECTION
from hwt.hdlObjects.typeShortcuts import vecT
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BIT
from hwt.interfaces.signalOps import SignalOps
from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.vectorUtils import getWidthExpr


D = DIRECTION


class Signal(SignalOps, Interface):
    def __init__(self,
                 masterDir=D.OUT,
                 multipliedBy=None,
                 dtype=BIT,
                 alternativeNames=None,
                 loadConfig=True):
        super().__init__(masterDir=masterDir,
                         multipliedBy=multipliedBy,
                         alternativeNames=alternativeNames,
                         loadConfig=loadConfig)
        self._dtype = dtype
        self._setMultipliedBy(multipliedBy, updateTypes=True)

    def _injectMultiplerToDtype(self):
        t = self._dtype
        factor = self._multipliedBy
        if t == BIT:
            newT = vecT(factor)
        elif isinstance(t, Bits):
            w = getWidthExpr(t)
            if isinstance(w, RtlSignalBase):
                # bouth Param or factor Value
                newW = w * factor
            elif isinstance(factor, RtlSignalBase):
                # w is Value
                newW = factor * w
            else:
                # bouth Value
                newW = w.clone()
                newW.val *= factor.val
            newT = vecT(newW)
        else:
            raise NotImplementedError("type:%s" % (repr(t)))
        self._dtype = newT

    def _setMultipliedBy(self, factor, updateTypes=True):
        if type(self._multipliedBy) == type(factor) and self._multipliedBy == factor:
            pass
        else:
            self._multipliedBy = factor
            if updateTypes and factor is not None:
                self._injectMultiplerToDtype()

    def _getSimAgent(self):
        from hwt.interfaces.agents.signal import SignalAgent
        return SignalAgent


def VectSignal(width, signed=None, masterDir=D.OUT, multipliedBy=None,
                   alternativeNames=None,
                   loadConfig=True):
    return Signal(masterDir, multipliedBy, vecT(width, signed), 
                   alternativeNames, loadConfig)        


class Clk(Signal):
    _alternativeNames = ['ap_clk', 'aclk', 'clk', 'clock']

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Clk
        return IP_Clk

    def _getSimAgent(self):
        from hwt.interfaces.agents.clk import OscilatorAgent
        return OscilatorAgent


class Rst(Signal):
    _alternativeNames = ['ap_rst', 'areset', 'reset', 'rst']

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst
        return IP_Rst

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullDownAgent
        return PullDownAgent


class Rst_n(Signal):
    _alternativeNames = ['ap_rst_n', 'aresetn', 'resetn', 'rstn', 'rst_n']

    def _signalsForInterface(self,
                             context,
                             prefix='',
                             typeTransform=None):
        sigs = Signal._signalsForInterface(self,
                                           context,
                                           prefix,
                                           typeTransform=typeTransform)
        for s in sigs:
            s.negated = True
        return sigs

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst_n
        return IP_Rst_n

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullUpAgent
        return PullUpAgent


class VldSynced(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.vld = s(alternativeNames=['valid'])

    def _getSimAgent(self):
        from hwt.interfaces.agents.vldSynced import VldSyncedAgent
        return VldSyncedAgent


class RdSynced(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.rd = s(masterDir=D.IN, alternativeNames=['ready'])

    def _getSimAgent(self):
        from hwt.interfaces.agents.rdSynced import RdSyncedAgent
        return RdSyncedAgent


class Handshaked(VldSynced):
    def _declr(self):
        super()._declr()
        self.rd = s(masterDir=D.IN, alternativeNames=['ready'])

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakedAgent
        return HandshakedAgent


class HandshakeSync(Interface):
    def _declr(self):
        self.vld = s(alternativeNames=['valid'])
        self.rd = s(masterDir=D.IN, alternativeNames=['ready'])

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakeSyncAgent
        return HandshakeSyncAgent


class ReqDoneSync(Interface):
    def _declr(self):
        self.req = s()
        self.done = s(masterDir=D.IN)


class BramPort_withoutClk(Interface):
    def _config(self):
        self.ADDR_WIDTH = Param(32)
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.addr = s(dtype=vecT(self.ADDR_WIDTH), alternativeNames=['addr_v'])
        self.din = s(dtype=vecT(self.DATA_WIDTH), alternativeNames=['din_v'])
        self.dout = s(masterDir=D.IN, dtype=vecT(self.DATA_WIDTH), alternativeNames=['dout_v'])
        self.en = s()
        self.we = s()

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_BlockRamPort
        return IP_BlockRamPort

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPort_withoutClkAgent
        return BramPort_withoutClkAgent


class BramPort(BramPort_withoutClk):

    def _declr(self):
        super()._declr()
        self.clk = s(masterDir=D.OUT)

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPortAgent
        return BramPortAgent

    @classmethod
    def fromBramPort_withoutClk(cls, bramPort, clk):
        assert isinstance(bramPort, BramPort_withoutClk)
        assert isinstance(clk, Clk)
        self = cls()
        rp = self._replaceParam

        def setIntf(name, intf):
            setattr(self, name, intf)
            self._interfaces.append(intf)

        rp("ADDR_WIDTH", bramPort.ADDR_WIDTH)
        rp("DATA_WIDTH", bramPort.DATA_WIDTH)

        self._interfaces = []
        for iName in ["addr", "din", "dout", "en", "we"]:
            intf = getattr(bramPort, iName)
            setIntf(iName, intf)
        setIntf("clk", clk)

        return self


class FifoWriter(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.en = s()
        self.wait = s(masterDir=D.IN)
        self.data = s(dtype=vecT(self.DATA_WIDTH), alternativeNames=[''])

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoWriterAgent
        return FifoWriterAgent


class FifoReader(FifoWriter):
    def _declr(self):
        super()._declr()
        self.en._masterDir = D.IN
        self.wait._masterDir = D.OUT

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoReaderAgent
        return FifoReaderAgent


class RegCntrl(Interface):
    """
    Register control interface
    """
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.din = Signal(dtype=vecT(self.DATA_WIDTH), masterDir=D.IN)
        with self._paramsShared():
            self.dout = VldSynced()

    def _getSimAgent(self):
        from hwt.interfaces.agents.regCntrl import RegCntrlAgent
        return RegCntrlAgent

s = Signal



from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.defs import BIT
from hwt.hdlObjects.types.hdlType import HdlType

def convertBoolean(self, sigOrVal, toType):
    if toType == BIT:
        if isinstance(sigOrVal, Value):
            v = BIT.getValueCls()(int(sigOrVal.val), BIT, sigOrVal.vldMask, sigOrVal.updateTime)
            return v 
        else:
            return sigOrVal._ternary(BIT.fromPy(1), BIT.fromPy(0))
        
    return HdlType.defaultConvert(self, sigOrVal, toType)



from hwt.hdlObjects.statements import ReturnCalled

class Function():
    """
    Container for hdl cunction
    """
    def __init__(self, name, returnT, ctx, params, _locals, exprList, isOperator=False):
        """
        class to store hdl function

        @param name: name of the function
        @param returnT: return type
        @param ctx: hdl context of this funtion
        @param params: list of argument signals
        @param _locals : local variables in this function
        @param exprList: list of expressions in body
        @param isOperator: is operator flag
        """
        self.name = name
        self.returnT = returnT
        self.ctx = ctx
        self.params = params
        self.locals = _locals
        self.exprList = exprList
        self.isOperator = isOperator

    def call(self, *args):
        for p, a in zip(self.params, args):
            a.updateTime = 0
            p._val = a
        
        try:
            for s in self.exprList:
                s.seqEval()
        except ReturnCalled as r:
            return r.val
        return None
    
    def staticEval(self):
        return self


import opcode

class Opcode(int):__str__ = __repr__ = lambda s:opcode.opname[s]

opmap = dict((name.replace('+', '_'), Opcode(code)) for name, code in opcode.opmap.items())
    
    
POP_TOP = 1
ROT_TWO = 2
ROT_THREE = 3
DUP_TOP = 4
DUP_TOP_TWO = 5
NOP = 9
UNARY_POSITIVE = 10
UNARY_NEGATIVE = 11
UNARY_NOT = 12
UNARY_INVERT = 15
BINARY_POWER = 19
BINARY_MULTIPLY = 20
BINARY_MODULO = 22
BINARY_ADD = 23
BINARY_SUBTRACT = 24
BINARY_SUBSCR = 25
BINARY_FLOOR_DIVIDE = 26
BINARY_TRUE_DIVIDE = 27
INPLACE_FLOOR_DIVIDE = 28
INPLACE_TRUE_DIVIDE = 29
STORE_MAP = 54
INPLACE_ADD = 55
INPLACE_SUBTRACT = 56
INPLACE_MULTIPLY = 57
INPLACE_MODULO = 59
STORE_SUBSCR = 60
DELETE_SUBSCR = 61
BINARY_LSHIFT = 62
BINARY_RSHIFT = 63
BINARY_AND = 64
BINARY_XOR = 65
BINARY_OR = 66
INPLACE_POWER = 67
GET_ITER = 68
PRINT_EXPR = 70
LOAD_BUILD_CLASS = 71
YIELD_FROM = 72
INPLACE_LSHIFT = 75
INPLACE_RSHIFT = 76
INPLACE_AND = 77
INPLACE_XOR = 78
INPLACE_OR = 79
BREAK_LOOP = 80
WITH_CLEANUP = 81
RETURN_VALUE = 83
IMPORT_STAR = 84
YIELD_VALUE = 86
POP_BLOCK = 87
END_FINALLY = 88
POP_EXCEPT = 89
STORE_NAME = 90
DELETE_NAME = 91
UNPACK_SEQUENCE = 92
FOR_ITER = 93
UNPACK_EX = 94
STORE_ATTR = 95
DELETE_ATTR = 96
STORE_GLOBAL = 97
DELETE_GLOBAL = 98
LOAD_CONST = 100
LOAD_NAME = 101
BUILD_TUPLE = 102
BUILD_LIST = 103
BUILD_SET = 104
BUILD_MAP = 105
LOAD_ATTR = 106
COMPARE_OP = 107
IMPORT_NAME = 108
IMPORT_FROM = 109
JUMP_FORWARD = 110
JUMP_IF_FALSE_OR_POP = 111
JUMP_IF_TRUE_OR_POP = 112
JUMP_ABSOLUTE = 113
POP_JUMP_IF_FALSE = 114
POP_JUMP_IF_TRUE = 115
LOAD_GLOBAL = 116
CONTINUE_LOOP = 119
SETUP_LOOP = 120
SETUP_EXCEPT = 121
SETUP_FINALLY = 122
LOAD_FAST = 124
STORE_FAST = 125
DELETE_FAST = 126
RAISE_VARARGS = 130
CALL_FUNCTION = 131
MAKE_FUNCTION = 132
BUILD_SLICE = 133
MAKE_CLOSURE = 134
LOAD_CLOSURE = 135
LOAD_DEREF = 136
STORE_DEREF = 137
DELETE_DEREF = 138
CALL_FUNCTION_VAR = 140
CALL_FUNCTION_KW = 141
CALL_FUNCTION_VAR_KW = 142
SETUP_WITH = 143
LIST_APPEND = 145
SET_ADD = 146
MAP_ADD = 147
LOAD_CLASSDEREF = 148
EXTENDED_ARG = 144


    
    
opcodes = set(Opcode(x) for x in  opcode.opmap.values())
cmp_op = opcode.cmp_op
hasarg = set(x for x in opcodes if x >= opcode.HAVE_ARGUMENT)
hasconst = set(Opcode(x) for x in opcode.hasconst)
hasname = set(Opcode(x) for x in opcode.hasname)
hasjrel = set(Opcode(x) for x in opcode.hasjrel)
hasjabs = set(Opcode(x) for x in opcode.hasjabs)
hasjump = hasjabs | hasjrel
haslocal = set(Opcode(x) for x in opcode.haslocal)
hascompare = set(Opcode(x) for x in opcode.hascompare)
hasfree = set(Opcode(x) for x in opcode.hasfree)
hascode = set((MAKE_FUNCTION, MAKE_CLOSURE))

_se = {
    IMPORT_FROM:1,
    DUP_TOP:1,
    LOAD_CONST:1,
    LOAD_NAME:1,
    LOAD_GLOBAL:1,
    LOAD_FAST:1,
    LOAD_CLOSURE:1,
    LOAD_DEREF:1,
    BUILD_MAP:1,
    
    YIELD_VALUE:0,
    UNARY_POSITIVE:0,
    UNARY_NEGATIVE:0,
    UNARY_NOT:0,
    UNARY_INVERT:0,
    GET_ITER:0,
    LOAD_ATTR:0,
    IMPORT_NAME:-1,
    ROT_TWO:0,
    ROT_THREE:0,
    NOP:0,
    DELETE_GLOBAL:0,
    DELETE_NAME:0,
    DELETE_FAST:0,
    
    IMPORT_NAME:-1,
    POP_TOP:-1,
    PRINT_EXPR:-1,
    IMPORT_STAR:-1,
    DELETE_ATTR:-1,
    STORE_DEREF:-1,
    STORE_NAME:-1,
    STORE_GLOBAL:-1,
    STORE_FAST:-1,
    BINARY_POWER:-1,
    BINARY_MULTIPLY:-1,
    BINARY_FLOOR_DIVIDE:-1,
    BINARY_TRUE_DIVIDE:-1,
    BINARY_MODULO:-1,
    BINARY_ADD:-1,
    BINARY_SUBTRACT:-1,
    BINARY_SUBSCR:-1,
    BINARY_LSHIFT:-1,
    BINARY_RSHIFT:-1,
    BINARY_AND:-1,
    BINARY_XOR:-1,
    BINARY_OR:-1,
    COMPARE_OP:-1,
    INPLACE_POWER:-1,
    INPLACE_MULTIPLY:-1,
    INPLACE_FLOOR_DIVIDE:-1,
    INPLACE_TRUE_DIVIDE:-1,
    INPLACE_MODULO:-1,
    INPLACE_ADD:-1,
    INPLACE_SUBTRACT:-1,
    INPLACE_LSHIFT:-1,
    INPLACE_RSHIFT:-1,
    INPLACE_AND:-1,
    INPLACE_XOR:-1,
    INPLACE_OR:-1,
    
    LIST_APPEND:-2,
    DELETE_SUBSCR:-2,
    STORE_ATTR:-2,
    STORE_SUBSCR:-3,
    }

_rf = {
    CALL_FUNCTION:lambda x:-((x & 0xFF00) >> 7) - (x & 0xFF),
    CALL_FUNCTION_VAR_KW:lambda x:-((x & 0xFF00) >> 7) - (x & 0xFF) - 2,
    CALL_FUNCTION_VAR:lambda x:-((x & 0xFF00) >> 7 | 1) - (x & 0xFF),
    CALL_FUNCTION_KW:lambda x:-((x & 0xFF00) >> 7 | 1) - (x & 0xFF),
   
    DUP_TOP:lambda x:x,
    RAISE_VARARGS:lambda x:x,
    MAKE_FUNCTION:lambda x:x,
    UNPACK_SEQUENCE:lambda x:x - 1,
    MAKE_CLOSURE:lambda x:x - 1,
    BUILD_TUPLE:lambda x:1 - x,
    BUILD_LIST:lambda x:1 - x,
    BUILD_SLICE:lambda x:1 - x}

hasflow = opcodes - set(_se) - set(_rf)




class HdlSimConfig():
    """
    Container of configuration of hdl simulator
    """
        
    def beforeSim(self, simulator, signals):
        """
        called beforee preparing of simulation
        """
        pass
    
    def logChange(self, nowTime, sig, nextVal):
        """
        Log change of value for signal
        """
        pass
        
    def logPropagation(self, x):
        """
        Log value propagation over netlist
        """
        pass

    def logApplyingValues(self, simulator, values):
        """
        Log simulator value quantum applied
        """
        pass


from hwt.hdlObjects.types.hdlType import HdlType 

class String(HdlType):
    def __init__(self):
        super().__init__()
        self.name = "STRING"

    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.stringConversions import convertString
        return convertString
    

    def all_mask(self):
        return 1

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.stringVal import StringVal 
            cls._valCls = StringVal
            return cls._valCls


from hwt.hdlObjects.constants import Unconstrained
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.serializer.exceptions import SerializerException


class VhdlSerializer_types():
    @classmethod
    def HdlType_bits(cls, typ, createTmpVarFn, declaration=False):
        disableRange = False
        if typ.signed is None:
            if typ.forceVector or typ.bit_length() > 1:
                name = 'STD_LOGIC_VECTOR'
            else:
                name = 'STD_LOGIC'
                disableRange = True
        elif typ.signed:
            name = "SIGNED"
        else:
            name = 'UNSIGNED'     
            
        c = typ.constrain
        if disableRange or c is None or isinstance(c, Unconstrained):
            constr = ""
        elif isinstance(c, (int, float)):
            constr = "(%d DOWNTO 0)" % (c - 1)
        else:        
            constr = "(%s)" % cls.Value(c, createTmpVarFn)     
        return name + constr


    @classmethod
    def HdlType_enum(cls, typ, scope, declaration=False):
        buff = []
        if declaration:
            try:
                name = typ.name
            except AttributeError:
                name = "enumT_"
            typ.name = scope.checkedName(name, typ)
            
            buff.extend(["TYPE ", typ.name.upper(), ' IS ('])
            # [TODO] check enum values names 
            buff.append(", ".join(typ._allValues))
            buff.append(")")
            return "".join(buff)
        else:
            return typ.name
        
    @classmethod
    def HdlType_array(cls, typ, createTmpVarFn, scope, declaration=False):
        if declaration:
            try:
                name = typ.name
            except AttributeError:
                name = "arrT_"
            
            typ.name = scope.checkedName(name, typ)
            
            return "TYPE %s IS ARRAY ((%s) DOWNTO 0) OF %s" % \
                (typ.name, cls.asHdl(toHVal(typ.size) - 1, createTmpVarFn), cls.HdlType(typ.elmType, createTmpVarFn))
        else:
            try:
                return typ.name
            except AttributeError:
                # [TODO]
                # sometimes we need to debug expression and we need temporary type name
                # this may be risk and this should be done by extra debug serializer
                return "arrT_%d" % id(typ) 
    
    @classmethod
    def HdlType_int(cls, typ, scope, declaration=False):
        ma = typ.max
        mi = typ.min
        noMax = ma is None
        noMin = mi is None
        if noMin: 
            if noMax:
                return "INTEGER"
            else:
                raise SerializerException("If max is specified min has to be specified as well")
        else:
            if noMax:
                if mi == 0:
                    return "NATURAL"
                elif mi == 1:
                    return "POSITIVE"
                else:
                    raise SerializerException("If max is specified min has to be specified as well")
            else:
                return "INTEGER RANGE %d to %d" % (mi, ma)
    
    @classmethod
    def HdlType(cls, typ, createTmpVarFn, scope=None, declaration=False):
        if isinstance(typ, Bits):
            return cls.HdlType_bits(typ, createTmpVarFn, declaration=declaration)
        elif isinstance(typ, Enum):
            return cls.HdlType_enum(typ, scope, declaration=declaration)
        elif isinstance(typ, Array):
            return cls.HdlType_array(typ, createTmpVarFn, scope, declaration=declaration)
        elif isinstance(typ, Integer):
            return cls.HdlType_int(typ, scope, declaration=declaration)
        else:
            if declaration:
                raise NotImplementedError("type declaration is not implemented for type %s" % 
                                      (typ.name))
            else:
                assert isinstance(typ, HdlType)
                return typ.name.upper()



from datetime import datetime
from pprint import pprint
import sys

from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.boolean import Boolean
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.simulator.vcdWritter import VcdWritter
from hwt.hdlObjects.types.enum import Enum


class VcdHdlSimConfig(HdlSimConfig):
    supported_type_classes = (Boolean, Bits, Enum)
    
    def __init__(self, dumpFile=sys.stdout):
        super().__init__()
        self.vcdWritter = VcdWritter(dumpFile) 
        self.logPropagation = False
        self.logApplyingValues = False        

        # unit :  signal | unit
        # signal : None
        self.registered = {}
    
    
    def logApplyingValues(self, simulator, values):
        pprint((simulator.now, values))
        
    def logPropagation(self, simulator, signal, process): 
        print("%d: Signal.simPropagateChanges %s -> %s" % 
                                        (simulator.now, signal.name, str(process.name))
        )
        
    def vcdRegisterUnit(self, unit):
        with self.vcdWritter.module(unit._name) as m:
            for se in unit._cntx.signals:
                if isinstance(se._dtype, self.supported_type_classes):
                    m.var(se)
                    
            for u in unit._units:
                self.vcdRegisterUnit(u)
   
    
    def _registerSignal(self, sig):
        self.registered[sig] = None
   
    def beforeSim(self, simulator, synthesisedUnit):
        """
        This method is called before first step of simulation.
        """
        self.vcdWritter.date(datetime.now())
        self.vcdWritter.timescale(1)

        self.vcdRegisterUnit(synthesisedUnit)
        self.vcdWritter.enddefinitions()
        
    def logChange(self, nowTime, sig, nextVal):
        """
        This method is called for every value change of any signal.
        """
        try:
            self.vcdWritter.change(nowTime, sig, nextVal)
        except KeyError:
            # not every signal has to be registered
            pass



from hwt.simulator.shortcuts import onRisingEdge
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase


class AgentBase():
    """
    Base class of agent of interface like in UVM
    driver is used for slave interfaces
    monitor is used for master interfaces

    @ivar intf: interface assigned to this agent
    @ivar enable: flag to enable/disable this agent
    """
    def __init__(self, intf):
        self.intf = intf
        self.enable = True
        self._debugOutput = None

    def _debug(self, out):
        self._debugOutput = out

    def getDrivers(self):
        """
        Called before simulation to collect all drivers of interfaces from this agent
        """
        return [self.driver]

    def getMonitors(self):
        """
        Called before simulation to collect all monitors of interfaces from this agent
        """
        return [self.monitor]

    def driver(self, s):
        """
        Implement this method to drive your interface in simulation/verification
        """
        raise NotImplementedError()

    def monitor(self, s):
        """
        Implement this method to monitor your interface in simulation/verification
        """
        raise NotImplementedError()


class SyncAgentBase(AgentBase):
    """
    Agent which discovers clk, rst signal and runs only at specified edge of clk
    @attention: requires clk and rst/rstn signal
      (if you do not have any create simulation wrapper with it)
    """
    def __init__(self, intf, clk=None, rstn=None, allowNoReset=False):
        super().__init__(intf)

        # resolve clk and rstn
        if clk is None:
            self.clk = self._getClk()
        else:
            self.clk = clk

        if rstn is None:
            self.rst_n = self.getRst_n(intf._parent, allowNoReset=allowNoReset)
        else:
            self.rst_n = rstn

        self.monitor = onRisingEdge(self.clk, self.monitor)
        self.driver = onRisingEdge(self.clk, self.driver)

    def getRst_n(self, parent, allowNoReset):
        while True:
            try:
                return parent.rst_n
            except AttributeError:
                pass

            try:
                return ~parent.rst
            except AttributeError:
                pass

            if isinstance(parent, UnitBase):
                break
            else:
                parent = parent._parent

        if allowNoReset:
            return None
        else:
            raise Exception("Can not find reset on unit %s" % (repr(parent)))

    def notReset(self, s):
        if self.rst_n is None:
            return True
        else:
            return s.r(self.rst_n).val

    def _getClk(self):
        p = self.intf._parent
        while True:
            try:
                return p.clk
            except AttributeError:
                if isinstance(p, UnitBase):
                    raise Exception("Can not find clk")
                p = p._parent
        return None




class TypeConversionErr(Exception):
    pass

class ConfErr(Exception):
    pass

class IntfLvlConfErr(ConfErr):
    """
    Interface level synthesizer user configuration error
    """
    pass

class SigLvlConfErr(ConfErr):
    """
    Signal level synthesizer user configuration error
    """
    pass



def areSameSignals(a, b):
    if a is b:
        return True
    if type(a) != type(b):
        return False 
    if len(a.drivers) != 1 or len(b.drivers) != 1:
        return False
    da = list(a.drivers)[0]
    db = list(b.drivers)[0]
    return da == db


from hwt.interfaces.std import  BramPort, \
                 BramPort_withoutClk, Handshaked, Clk, Rst_n, Signal, Rst, RdSynced, VldSynced
#from hwtLib.amba.amba import Axi4, Axi4_xil, \
#    AxiLite, AxiLite_xil, AxiStream, AxiStream_withoutSTRB, AxiStream_withUserAndStrb, \
#    AxiStream_withUserAndNoStrb 

allInterfaces = [#Axi4, Axi4_xil,
   #AxiLite, AxiLite_xil,
   BramPort, BramPort_withoutClk,
   #AxiStream_withUserAndStrb, AxiStream, AxiStream_withUserAndNoStrb, AxiStream_withoutSTRB,
   Handshaked,  Clk, Rst, Rst_n, VldSynced, RdSynced, Signal
 ]



from hwt.synthesizer.exceptions import TypeConversionErr


class InvalidVHDLTypeExc(Exception):
    def __init__(self, vhdlType):
        self.vhdlType = vhdlType
    
    def __str__(self):    
        variableName = self.variable.name
        return "Invalid type, width is %s in the context of variable %s" \
            % (str(self.vhdlType.getWidth()), variableName)

    def __repr__(self):
        return self.__str__()

class HdlType():
    #_slots__ = ['name', "constrain", "_valCls", "_convertor"]

    def __init__(self):
        self.constrain = None
        
    def __eq__(self, other):
        return type(self) is type(other)
    
    def __hash__(self):
        return hash((self.name, self.constrain))
    
    def fromPy(self, v):
        return self.getValueCls().fromPy(v, self)
    
    def convert(self, sigOrVal, toType):
        if sigOrVal._dtype == toType:
            return sigOrVal
        
        try:
            c = self._convert
        except AttributeError:
            c = self.getConvertor()
            self._convertor = c

        return c(self, sigOrVal, toType)
    
    @classmethod
    def getConvertor(cls):
        return HdlType.defaultConvert
    
    def defaultConvert(self, sigOrVal, toType):
        raise TypeConversionErr("Conversion of %r of type \n%r to type %r is not implemented" 
                                   % (sigOrVal, self, toType))
    
    def __repr__(self):
        return "<HdlType %s>" % (self.__class__.__name__)



from enum import Enum

# [TODO] rename to constants

class Unconstrained():
    __slots__ = ["derivedWidth"]
    
class INTF_DIRECTION(Enum):
    MASTER = 0
    SLAVE = 1
    TRISTATE = 2
    UNKNOWN = 3
    
    @classmethod
    def asDirection(cls, val):
        if val == INTF_DIRECTION.SLAVE:
            return DIRECTION.IN 
        elif val == INTF_DIRECTION.MASTER:
            return DIRECTION.OUT
        elif val == INTF_DIRECTION.TRISTATE:
            return DIRECTION.INOUT
        else:
            raise Exception("Parameter (%s) is not interface direction" % (repr(val)))
    
    @classmethod
    def opposite(cls, d):
        if d == cls.SLAVE:
            return cls.MASTER
        elif d == cls.MASTER:
            return cls.SLAVE
        elif d == cls.TRISTATE:
            return d
        else:
            raise Exception("%s is not interface direction" % (repr(d)))
    
READ = "READ"
WRITE = "WRITE"
NOP = "NOP"
    
class DIRECTION(Enum):
    IN = 0
    OUT = 1
    INOUT = 2
    
    @classmethod
    def asIntfDirection(cls, d):
        if d == cls.IN:
            return INTF_DIRECTION.SLAVE
        elif d == cls.OUT:
            return INTF_DIRECTION.MASTER
        elif d == cls.INOUT:
            return INTF_DIRECTION.TRISTATE
        else:
            raise TypeError("Parameter %s is not direction" % (str(d)))
    
    @classmethod
    def opposite(cls, d):
        if d == cls.IN:
            return cls.OUT
        elif d == cls.OUT:
            return cls.IN
        elif d == cls.INOUT:
            return d
        else:
            raise Exception("Parameter is not direction")
            

class Time():
    ps = 1
    ns = 1000
    us = ns * 1000
    ms = us * 1000
    s = ms * 1000

class SENSITIVITY(Enum):
    ANY = 0b11
    RISING = 0b01 
    FALLING = 0b10 



from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.integer import Integer 

class Slice(Array):
    def __init__(self):
        super().__init__(Integer(), 2)
        
    def __hash__(self):
        return hash(self.constrain)

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.sliceVal import SliceVal 
            cls._valCls = SliceVal
            return cls._valCls


from hwt.bitmask import mask
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.slice import Slice
from hwt.hdlObjects.types.string import String
from hwt.hdlObjects.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class VhdlSerializer_Value():
    
    @classmethod
    def Value(cls, val, createTmpVarFn):
        """ 
        @param dst: is signal connected with value 
        @param val: value object, can be instance of Signal or Value    """
        t = val._dtype
        if isinstance(val, RtlSignalBase):
            return cls.SignalItem(val, createTmpVarFn)
        elif isinstance(t, Slice):
            return cls.Slice_valAsVhdl(t, val, createTmpVarFn)
        elif isinstance(t, Array):
            return cls.Array_valAsVhdl(t, val, createTmpVarFn)
        elif isinstance(t, Bits):
            return cls.Bits_valAsVhdl(t, val)
        elif isinstance(t, Boolean):
            return cls.Bool_valAsVhdl(t, val)
        elif isinstance(t, Enum):
            return cls.Enum_valAsVhdl(t, val)
        elif isinstance(t, Integer):
            return cls.Integer_valAsVhdl(t, val)
        elif isinstance(t, String):
            return cls.String_valAsVhdl(t, val)
        else:
            raise Exception("value2vhdlformat can not resolve value serialization for %s" % (repr(val))) 
    
    @classmethod
    def SignalItem(cls, si, createTmpVarFn, declaration=False):
        if declaration:
            v = si.defaultVal
            if si.virtualOnly:
                prefix = "VARIABLE"
            elif si.drivers:
                prefix = "SIGNAL"
            elif si.endpoints or si.simSensProcs:
                prefix = "CONSTANT"
                if not v.vldMask:
                    raise SerializerException("Signal %s is constant and has undefined value" % si.name)
            else:
                raise SerializerException("Signal %s should be declared but it is not used" % si.name)
                

            s = prefix + " %s : %s" % (si.name, cls.HdlType(si._dtype, createTmpVarFn))
            if isinstance(v, RtlSignalBase):
                return s + " := %s" % cls.asHdl(v, createTmpVarFn)
            elif isinstance(v, Value):
                if si.defaultVal.vldMask:
                    return s + " := %s" % cls.Value(si.defaultVal, createTmpVarFn)
                else:
                    return s 
            else:
                raise NotImplementedError(v)
            
        else:
            if si.hidden and hasattr(si, "origin"):
                return cls.asHdl(si.origin, createTmpVarFn)
            else:
                return si.name

    @classmethod
    def Enum_valAsVhdl(cls, dtype, val):
        return  '%s' % str(val.val)
    
    @classmethod
    def Array_valAsVhdl(cls, dtype, val, createTmpVarFn):
        return  "(" + (",\n".join([cls.Value(v, createTmpVarFn) for v in val.val])) + ")"
    
        
    @classmethod
    def Bits_valAsVhdl(cls, dtype, val):
        w = dtype.bit_length()
        if dtype.signed is None:
            if dtype.forceVector or w > 1:
                return cls.BitString(val.val, w, val.vldMask)
            else:
                return cls.BitLiteral(val.val, val.vldMask)
        elif dtype.signed:
            return cls.SignedBitString(val.val, w, dtype.forceVector, val.vldMask)
        else:
            return cls.UnsignedBitString(val.val, w, dtype.forceVector, val.vldMask)

    @staticmethod
    def BitString_binary(v, width, vldMask=None):
        buff = []
        for i in range(width - 1, -1, -1):
            mask = (1 << i)
            b = v & mask
            
            if vldMask & mask:
                s = "1" if b else "0"
            else:
                s = "X"
            buff.append(s)
        return '"%s"' % (''.join(buff))
    
    @classmethod
    def BitString(cls, v, width, vldMask=None):
        if vldMask is None:
            vldMask = mask(width)
        # if can be in hex
        if width % 4 == 0 and vldMask == (1 << width) - 1:
            return ('X"%0' + str(width // 4) + 'x"') % (v)
        else:  # else in binary
            return cls.BitString_binary(v, width, vldMask)
    
    @classmethod
    def BitLiteral(cls, v, vldMask):
        if vldMask:
            return  "'%d'" % int(bool(v))
        else:
            return "'X'"

    
    @classmethod
    def SignedBitString(cls, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        return "TO_SIGNED(%s, %d)" % (v, width)

    
    @classmethod
    def UnsignedBitString(cls, v, width, forceVector, vldMask):
        if vldMask != mask(width):
            if forceVector or width > 1:
                v = cls.BitString(v, width, vldMask)
            else:
                v = cls.BitLiteral(v, width, vldMask)
        else:
            v = str(v)
        # [TODO] parametrized width
        return "TO_UNSIGNED(%s, %d)" % (v, width)
    
    @classmethod
    def Bool_valAsVhdl(cls, dtype, val):
        return str(bool(val.val))
    
    @classmethod
    def Integer_valAsVhdl(cls, dtype, val):
        return str(int(val.val))

    @classmethod
    def Slice_valAsVhdl(cls, dtype, val, createTmpVarFn):
        return "%s DOWNTO %s" % (cls.Value(val.val[0], createTmpVarFn), cls.Value(val.val[1], createTmpVarFn))

    @classmethod
    def String_valAsVhdl(cls, dtype, val):
        return  '"%s"' % str(val.val)
    




from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import Integer, BIT
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.sliceVal import SliceVal
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.typeShortcuts import vec

def getWidthExpr(vectorTypeInst):
    c = vectorTypeInst.constrain
    if isinstance(c, SliceVal):
        return c.val[0] + 1
    downto = c.singleDriver()
    
    assert downto.operator == AllOps.DOWNTO
    assert downto.ops[1].val == 0
    
    
    widthMinOne = downto.ops[0]
    if isinstance(widthMinOne, Value) and isinstance(widthMinOne._dtype, Integer):
        w = widthMinOne.clone()
        w.val += 1
        return w
    else:
        widthMinOne = widthMinOne.singleDriver()
    assert widthMinOne.operator == AllOps.SUB
    assert widthMinOne.ops[1].val == 1
    
    return widthMinOne.ops[0]

def aplyIndexOnSignal(sig, dstType, index):
    if sig._dtype == BIT or dstType == BIT:
        return sig[index]
    elif isinstance(dstType, Bits):
        w = getWidthExpr(dstType)
        return sig[(w * (index + 1)):(w * index)]
    else:
        raise NotImplementedError()

def fitTo(what, to):
    """
    Slice signal "what" to fit in "to" 
    or
    extend "what" with zeros to same width as "to"
    
    little-endian impl.
    """

    whatWidth = what._dtype.bit_length()
    toWidth = to._dtype.bit_length()
    if toWidth == whatWidth:
        return what
    elif toWidth < whatWidth:
        # slice
        return what[toWidth:]
    else:
        # extend
        return vec(0, toWidth - whatWidth)._concat(what)


from hwt.simulator.agentBase import SyncAgentBase
from hwt.hdlObjects.constants import NOP


class VldSyncedAgent(SyncAgentBase):
    def __init__(self, intf, clk=None, rstn=None, allowNoReset=False):
        super(VldSyncedAgent, self).__init__(intf, clk=clk, rstn=rstn, allowNoReset=allowNoReset)
        self.data = []

    def doRead(self, s):
        return s.read(self.intf.data)

    def doWrite(self, s, data):
        s.w(data, self.intf.data)

    def monitor(self, s):
        intf = self.intf
        yield s.updateComplete
        if self.enable and self.notReset(s) and s.r(intf.vld).val:
            d = self.doRead(s)

            if self._debugOutput is not None:
                self._debugOutput.write("%s, read, %d: %r\n" % (
                                          self.intf._getFullName(), s.now, d))
            self.data.append(d)

    def driver(self, s):
        intf = self.intf

        if self.enable and self.data and self.notReset(s):
            d = self.data.pop(0)
            if d is NOP:
                self.doWrite(s, None)
                s.w(0, intf.vld)
            else:
                self.doWrite(s, d)
                s.w(1, intf.vld)
                if self._debugOutput is not None:
                    self._debugOutput.write("%s, wrote, %d: %r\n" % (
                                                self.intf._getFullName(), s.now, self.actualData))

        else:
            self.doWrite(s, None)
            s.w(0, intf.vld)



from hwt.bitmask import mask
from hwt.hdlObjects.types.bits import Bits
from hwt.simulator.types.simBitsConversions import convertSimBits__val


__simBitsTCache = {}
def simBitsT(width, signed):
    """
    Construct SimBitsT with cache
    """
    k = (width, signed)
    try:
        return __simBitsTCache[k] 
    except KeyError:
        t = SimBitsT(width, signed)
        __simBitsTCache[k] = t
        return t
    


class SimBitsT(Bits):
    """
    Simplified Bits type for simulation purposes
    """
    def __init__(self, widthConstr, signed):
        self.constrain = widthConstr
        self.signed = signed
        self._allMask = mask(self.bit_length())
    
    def __eq__(self, other):
        return isinstance(other, Bits) and other.bit_length() == self.bit_length()\
            and self.signed == other.signed
    
    def __hash__(self):
        return hash((self.constrain, self.signed))
    
    def all_mask(self):
        return self._allMask
    
    def bit_length(self):
        return self.constrain

    def convert(self, sigOrVal, toType):
        return convertSimBits__val(self, sigOrVal, toType)
    
SIM_BIT = SimBitsT(1, None)



class Assignment(object):
    """
    Assignment container
    @ivar src: source  
    @ivar dest: destination signal
    @ivar cond: set of terms if all them are evaluated to True,
                assignment is active
    @ivar indexes: description of index selector on dst (list of Index/Slice objects)
                    (f.e. [[0], [1]] means  dst[0][1]  )
    
    @cvar __instCntr: counter used for generating instance ids
    @ivar _instId: internaly used only for intuitive sorting of statements
    """
    __instCntr = 0
    
    def __init__(self, src, dst, indexes=None, virtualOnly=False):
        """
        @param dst: destination to assign to
        @param src: source which is assigned from
        @param indexes: description of index selector on dst (list of Index/Slice objects)
                (f.e. [[0], [1]] means  dst[0][1]  )
        @param virtualOnly: flag indicates that this assignments is only virtual and should not be added into
                netlist, because it is only for internal notation
        """
        self.src = src
        self.dst = dst
        self.isEventDependent = False
        self.indexes = indexes
        self.cond = set()
        self._instId = Assignment._nextInstId()
        
        if not virtualOnly:
            dst.ctx.startsOfDataPaths.add(self)
    
    @classmethod
    def _nextInstId(cls):
        """
        Get next instance id
        """
        i = cls.__instCntr
        cls.__instCntr += 1
        return i
        
    def seqEval(self):
        """Sequentially evaluate this assignment"""
        self.dst._val = self.src.staticEval() 
    
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        return VhdlSerializer.Assignment(self, onlyPrintDefaultValues)    




from hwt.hdlObjects.expr import ExprComparator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.constants import INTF_DIRECTION
from hwt.hdlObjects.types.sliceVal import SliceVal
from hwt.hdlObjects.value import Value
from hwt.synthesizer.interfaceLevel.interfaceUtils.array import InterfaceArray
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.walkers import walkSignalsInExpr
from hwt.pyUtils.arrayQuery import single, NoValueExc, arr_any
from hwt.pyUtils.stringUtils import matchIgnorecase


class InterfaceIncompatibilityExc(Exception):
    pass

def updateParam(intfParam, unitParam):
    if isinstance(unitParam, Param):
        intfParam.replace(unitParam)
    elif isinstance(unitParam, Value):
        intfParam.set(unitParam)
    else:
        # parameter resolution was not successful
        pass
    
def typeIsParametrized(dtype):
    c = dtype.constrain
    if c is None:
        return False
    else:
        return arr_any(list(walkSignalsInExpr(c)), lambda x: True)

class ExtractableInterface(InterfaceArray):

    @classmethod
    def _extractPossiblePrefixes(cls, ports, prefix=""):
        """
        @return: iterator over unit ports witch probably matches with this interface
        """
        intfObj = cls()
        intfObj._loadDeclarations()
        
        # find first signal in this interface
        parent = intfObj
        child = intfObj
        
        while child._interfaces:
            parent = child
            child = child._interfaces[0]
            # update prefix
            if child._interfaces:
                if prefix == '':
                    prefix = child._name
                else:
                    prefix += (parent._NAME_SEPARATOR + child._name)
            else:
                if prefix != '':
                    prefix += parent._NAME_SEPARATOR
                
        o = child
        name = child._name
            
        # search for alternative names as well    
        firstIntfNames = []
        for n in o._alternativeNames:
            firstIntfNames.append(prefix + n)
        firstIntfNames.append(prefix + name)
        
        # search ports names for firstIntfNames  
        for firstIntfName in firstIntfNames:
            for p in ports:
                if not hasattr(p, "_interface") and p.name.lower().endswith(firstIntfName):
                    # cut off prefix
                    nameLen = len(firstIntfName)
                    if nameLen == 0:
                        yield p.name
                    else:
                        yield p.name[:-nameLen]
                
    def _unExtrac(self):
        """undo extracting process for this interface"""
        if self._interfaces:
            for intfConfMap in self._interfaces:
                intfConfMap._unExtrac()
        else:
            if hasattr(self, "_boundedEntityPort"):
                if hasattr(self._boundedEntityPort, "_interface"):
                    del self._boundedEntityPort._interface
                del self._boundedEntityPort
                
    def _extractDtype(self, multipliedBy=None):
        """
        Compare signal _dtype and _dtype of interface declaration and try match parameters
        """
        self._multipliedBy = multipliedBy
        if self._interfaces:
            for si in self._interfaces:
                si._extractDtype(multipliedBy=multipliedBy)
        else:
            # update interface type from hdl, update generics
            intfTConstr = self._dtype.constrain

            if intfTConstr is not None and isinstance(intfTConstr, (RtlSignalBase, SliceVal)):
                unitTConstr = self._boundedEntityPort._dtype.constrain
                paramDiff = list(ExprComparator.findExprDiffInParam(intfTConstr, unitTConstr))
                    
                for intfParam, unitParam in paramDiff:
                    if multipliedBy is not None and isinstance(unitParam, RtlSignalBase):
                        mulOp = unitParam.singleDriver()
                        assert mulOp.operator == AllOps.MUL
                        op0 = mulOp.ops[0]
                        op1 = mulOp.ops[1]
                        
                        if type(op0) == type(multipliedBy) and op0 == multipliedBy:
                            _unitParam = op1
                        else:
                            assert type(op1) == type(multipliedBy) and op1 == multipliedBy
                            _unitParam = op0
                    else:
                        assert multipliedBy is None
                        _unitParam = unitParam 

                if len(paramDiff) == 0:
                    # [TODO] and port is not parametrized
                    origT = self._boundedEntityPort._dtype
                    t = self._dtype
                    self._dtypeMatch = origT == t and not typeIsParametrized(origT) and not typeIsParametrized(t) 
                else:
                    self._dtypeMatch = isinstance(_unitParam, (Param, Value))
                    updateParam(intfParam, _unitParam)
            else:
                self._dtypeMatch = self._boundedEntityPort._dtype == self._dtype
        
    def _tryToExtractByName(self, prefix, ports):
        """
        @return: self if extraction was successful
        @raise InterfaceIncompatibilityExc: if this interface with this prefix does not fit for this entity 
        """
        if self._interfaces:
            # extract subinterfaces and propagate params
            allDirMatch = True
            noneDirMatch = True
            if hasattr(self, "_name") and self._name != '':
                prefix += self._name + self._NAME_SEPARATOR
            try:
                for intf in self._interfaces:
                    intf._tryToExtractByName(prefix, ports)
                    if intf._interfaces:
                        dirMatches = intf._direction == INTF_DIRECTION.MASTER
                    else:
                        dirMatches = intf._boundedEntityPort.direction == intf._masterDir
                    allDirMatch = allDirMatch and dirMatches
                    noneDirMatch = noneDirMatch  and not dirMatches     
            except InterfaceIncompatibilityExc as e:
                for intf in self._interfaces:
                    intf._unExtrac()
                raise e
            
                
            if allDirMatch:
                self._direction = INTF_DIRECTION.MASTER
            elif noneDirMatch:
                self._direction = INTF_DIRECTION.SLAVE
            else:
                self._unExtrac()
                raise InterfaceIncompatibilityExc("Direction mismatch")
        
        else:
            # extract signal(Signal , etc.)
            # collect all possible names
            intfNames = []
            if hasattr(self, "_name"):
                intfNames.append(self._name)
            intfNames.extend(self._alternativeNames)
            # try find suitable portItem in entity port 
            for n in intfNames:
                name = prefix + n
                try:
                    self._boundedEntityPort = single(ports,
                                            lambda p : matchIgnorecase(p.name, name))
                    break
                except NoValueExc as e:
                    continue
            if not hasattr(self, "_boundedEntityPort"):
                self._unExtrac()
                raise  InterfaceIncompatibilityExc("Missing " + prefix + n)

            self._extractDtype()
            
            # assign references to hdl objects
            self._boundedEntityPort._interface = self
            # resolve direction
            dirMatches = self._boundedEntityPort.direction == self._masterDir
            if dirMatches:
                self._direction = INTF_DIRECTION.MASTER
            else:
                self._direction = INTF_DIRECTION.SLAVE

        return self
    
    @classmethod        
    def _tryToExtract(cls, ports):
        """
        @return: iterator over tuples (interface name. extracted interface)
        """
        # [TODO] ports as dict
        for name in cls._extractPossiblePrefixes(ports):
            try:
                intfInst = cls()
                intfInst._loadDeclarations()
                prefix = name 
                     
                intf = intfInst._tryToExtractByName(prefix, ports)
                if name == "":
                    if intf._interfaces:
                        name = cls.__name__.lower()
                    else:
                        name = intf._boundedEntityPort.name
                else:
                    name += intf._name
                
                if name.endswith("_"):
                    name = name[:-1]
                
                if intf._interfaces:  # if is possible to determine size of this potential intf. array 
                    # if interface is actually array of interfaces extract the size of this array 
                    mf = intf._tryExtractMultiplicationFactor()
                    if mf is not None:
                        intf._extractDtype(multipliedBy=mf)
                        intf._setMultipliedBy(mf, updateTypes=False)
                intfInst._setAsExtern(True)
                yield (name, intf) 
            except InterfaceIncompatibilityExc as e:
                # pass if interface was not found in ports
                pass




class ReturnCalled(Exception):
    """
    Exception which is used as return statement while executing of hdl functions
    """
    def __init__(self, val):
        self.val = val

class ReturnContainer():
    """
    Stuctural container of return statement in hdl
    """
    
    def __init__(self, val=None):
        self.val = val

    def seqEval(self):
        raise ReturnCalled(self.val.staticEval())       

def seqEvalCond(cond):
    _cond = True
    for c in cond:
        _cond = _cond and bool(c.staticEval().val)
        
    return _cond

class IfContainer():
    """
    Structural container of if statement for hdl rendering
    """
    def __init__(self, cond, ifTrue=[], ifFalse=[], elIfs=[]):
        self.cond = cond
        self.ifTrue = ifTrue
        self.elIfs = elIfs
        self.ifFalse = ifFalse
        
    def seqEval(self):
        if seqEvalCond(self.cond):
            for s in self.ifTrue:
                s.seqEval()
        else:
            for c in self.elIfs:
                if seqEvalCond(c[0]):
                    for s in c[1]:
                        s.seqEval()
                    return
            
            for s in self.ifFalse:
                s.seqEval()
        
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        return VhdlSerializer.IfContainer(self, onlyPrintDefaultValues)

class SwitchContainer():
    """
    Structural container for switch statement for hdl rendering
    """
    def __init__(self, switchOn, cases):
        self.switchOn = switchOn
        self.cases = cases
    
    def seqEval(self):
        raise NotImplementedError()
    
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        return VhdlSerializer.SwitchContainer(self, onlyPrintDefaultValues)
 
class WhileContainer():
    """
    Structural container of while statement for hdl rendering
    """
    def __init__(self, cond, body):
        self.cond = cond
        self.body = body
    
    def seqEval(self):
        while seqEvalCond(self.cond):
            for s in self.body:
                s.seqEval()

class WaitStm():
    """
    Structural container of wait statemnet for hdl rendering
    """
    def __init__(self, waitForWhat):
        self.isTimeWait = isinstance(waitForWhat, int)
        self.waitForWhat = waitForWhat
        



from hwt.hdlObjects.types.arrayVal import ArrayVal  
from hwt.hdlObjects.types.defs import SLICE
from hwt.hdlObjects.value import Value

class SliceVal(ArrayVal):

    def staticEval(self):
        _0 = self.val[0].staticEval()
        _1 = self.val[1].staticEval()
        updateTime = max(_0.updateTime, _1.updateTime)
        return SliceVal([_0, _1], SLICE, None, updateTime)
    
    def _isFullVld(self):
        return self.val[0]._isFullVld() and self.val[1]._isFullVld()

    def _size(self):
        assert isinstance(self, Value)
        
        return self.val[0].val - self.val[1].val +1


from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.typeCast import toHVal


class StringVal(Value):
    
    @classmethod
    def fromPy(cls, val, typeObj):
        assert isinstance(val, str) or val is None
        vld = 0 if val is None else 1
        if not vld:
            val = ""
        return cls(val, typeObj, vld)
    
    def _eq__val(self, other):
        eq = self.val == other.val
        vld = int(self.vldMask and other.vldMask)
        updateTime = max(self.updateTime, other.updateTime)
        
        return BOOL.getValueCls()(eq, BOOL, vld, updateTime)
        
    def _eq(self, other):
        other = toHVal(other)
        if isinstance(other, Value):
            return self._eq__val(other) 
        else:
            raise NotImplementedError()


from hwt.hdlObjects.constants import DIRECTION


class PortMap():
    def __init__(self, signal, portItem):
        self.sig = signal
        self.portItem = portItem
    
    @classmethod
    def fromPortItem(cls, portItem):
        d = portItem.direction
        if d == DIRECTION.IN:
            sig = portItem.src
        elif d == DIRECTION.OUT:
            sig = portItem.dst
        else:
            raise NotImplementedError()
        
        assert sig is not None
        
        return cls(sig, portItem)
    
    def __repr__(self):
        return "<PortMap %s => %s>" % (self.portItem.name, self.sig.name) 


from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdlObjects.operatorDefs import AllOps

opPrecedence = {AllOps.NOT : 4,
                AllOps.EVENT: 1,
                AllOps.RISING_EDGE: 1,
                AllOps.DIV: 4,
                AllOps.ADD : 5,
                AllOps.SUB: 5,
                AllOps.MUL: 4,
                AllOps.XOR: 9,
                AllOps.EQ: 10,
                AllOps.NEQ: 10,
                AllOps.AND_LOG: 10,
                AllOps.OR_LOG: 10,
                AllOps.DOWNTO: 1,
                AllOps.GREATERTHAN: 10,
                AllOps.LOWERTHAN: 10,
                AllOps.CONCAT: 1,
                AllOps.INDEX: 1,
                AllOps.TERNARY: 1,
                AllOps.CALL: 1,
                }

class SimModelSerializer_ops():
    @classmethod
    def BitToBool(cls, cast):
        v = 0 if cast.sig.negated else 1
        return cls.asHdl(cast.sig) + "._eq__val(hBit(%d))" % v

    @classmethod
    def Operator(cls, op):
        def p(operand):
            s = cls.asHdl(operand)
            if isinstance(operand, RtlSignalBase):
                try:
                    o = operand.singleDriver()
                    if opPrecedence[o.operator] <= opPrecedence[op.operator]:
                        return " (%s) " % s
                except Exception:
                    pass
            return " %s " % s
        
        ops = op.ops
        o = op.operator
        def _bin(name):
            return "(%s).%s(%s)" % (p(ops[0]).strip(), name, p(ops[1]).strip())
        
        if o == AllOps.AND_LOG:
            return _bin('_and__val')
        elif o == AllOps.OR_LOG:
            return _bin('_or__val')
        elif o == AllOps.XOR:
            return _bin('_xor__val')
        elif o == AllOps.NOT:
            assert len(ops) == 1
            return "(%s)._invert__val()" % p(ops[0])
        # elif o == AllOps.CALL:
        #    return "%s(%s)" % (cls.FunctionContainer(ops[0]), ", ".join(map(p, ops[1:])))
        elif o == AllOps.CONCAT:
            return "(%s)._concat__val(%s)" % (p(ops[0]), p(ops[1]))
        elif o == AllOps.DIV:
            return _bin('_floordiv__val')
        elif o == AllOps.DOWNTO:
            return "SliceVal((%s, %s), SLICE, True)" % (p(ops[0]), p(ops[1]))
        elif o == AllOps.EQ:
            return '(%s)._eq__val(%s)' % (p(ops[0]), p(ops[1]))
        elif o == AllOps.EVENT:
            assert len(ops) == 1
            return p(ops[0]) + "._hasEvent__val(sim)"
        elif o == AllOps.GREATERTHAN:
            return _bin('_gt__val')
        elif o == AllOps.GE:
            return _bin('_ge__val')
        elif o == AllOps.LE:
            return _bin('_le__val')
        elif o == AllOps.INDEX:
            assert len(ops) == 2
            return "(%s)._getitem__val(%s)" % ((cls.asHdl(ops[0])).strip(), p(ops[1]))
        elif o == AllOps.LOWERTHAN:
            return _bin('_lt__val')
        elif o == AllOps.SUB:
            return _bin('_sub__val')
        elif o == AllOps.MUL:
            return _bin('_mul__val')
        elif o == AllOps.NEQ:
            return _bin('_ne__val')
        elif o == AllOps.ADD:
            return _bin('_add__val')
        elif o == AllOps.TERNARY:
            return "(%s)._ternary__val(%s, %s)" % tuple(map(cls.asHdl, ops)) 
        elif o == AllOps.RISING_EDGE:
            assert len(ops) == 1
            return "(%s)._onRisingEdge__val(sim.now)" % (p(ops[0]))
        elif o == AllOps.FALLIGN_EDGE:
            assert len(ops) == 1
            return "(%s)._onFallingEdge__val(sim.now)" % (p(ops[0]))
        elif o == AllOps.BitsAsSigned:
            assert len(ops) == 1
            return  "(%s)._convSign__val(True)" % p(ops[0])
        elif o == AllOps.BitsAsUnsigned:
            assert len(ops) == 1
            return  "(%s)._convSign__val(False)" % p(ops[0])
        elif o == AllOps.BitsAsVec:
            assert len(ops) == 1
            return  "(%s)._convSign__val(None)" % p(ops[0])
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = ops[0]
            return "convertSimBits__val(%s, %s, SIM_INT)" % (cls.HdlType_bits(op._dtype), cls.asHdl(op))
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            return "convertSimInteger__val(%s, %s, %s)" % (cls.HdlType(ops[0]._dtype),
                                                   cls.asHdl(ops[0]),
                                                   cls.HdlType_bits(resT))
            
        elif o == AllOps.POW:
            assert len(ops) == 2
            return  "power(%s, %s)" % (p(ops[0]), p(ops[1]))
        else:
            raise NotImplementedError("Do not know how to convert %s to simModel" % (o))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

"""
Simple and stupid implementation of vhdl formater, no parser based on regular expressions
"""

indentIncr = ["^entity", "^port\s*\(", "^port\s*map\s*\(", "^generic\s*map\s*\(", "^generic\s*\(",
              "^architecture", "^if", "^case", "^port\s+map\s*\(", "^process", "^while", "^component",
              "\S+\s*:\s*process"]
indentDecr = ["^end", "^\)"]
indentPeak = ["^begin", "^elsif", "^else", "^when",]

indentIncr = list(map(lambda x: re.compile(x, re.IGNORECASE), indentIncr))
indentDecr = list(map(lambda x: re.compile(x, re.IGNORECASE), indentDecr))
indentPeak = list(map(lambda x: re.compile(x, re.IGNORECASE), indentPeak))


def get_indent(i):
    return "".join([" " for _ in range(i)])


def formatVhdl(vhdlString):
    indent = 0
    lines = []

    def getIndent(i):
        return get_indent(i * 4)

    for l in vhdlString.split("\n"):
        l = l.strip()
        if any([x.match(l) for x in indentDecr]):
            indent -= 1
            lines.append(getIndent(indent) + l)
        elif any([x.match(l) for x in indentIncr]):
            lines.append(getIndent(indent) + l)
            indent += 1
        elif any([x.match(l) for x in indentPeak]):
            lines.append(getIndent(indent - 1) + l)
        else:
            lines.append(getIndent(indent) + l)
    return "\n".join(lines)



from enum import Enum

class SERI_MODE(Enum):
    """
    Serializer mode for unit
    
    Modes are driving when unit should be serialized
    """
    
    # always serialize as independent unit
    ALWAYS = 0 
    
    # serialize only once and all other occurrences of this unit replace with first serialized one
    ONCE = 1 
    
    # serialize when params are unique, units with same params will be serialized as single unit
    PARAMS_UNIQ = 2 
    
    # exclude completely from serialization (useful for archetypes like LUT, GT, DPS block etc. which are already in vendor std. lib)
    # cls.__name__ (name of class) is used for all component instances
    EXCLUDE = 3


from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import INT, BOOL, STR, BIT
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.synthesizer.param import evalParam


# create hdl integer value (for example integer value in vhdl)
hInt = lambda val: INT.fromPy(val)

# create hdl boolean value (for example boolean value in vhdl)
hBool = lambda val: BOOL.fromPy(val)

# create hdl string value (for example string value in vhdl)
hStr = lambda val: STR.fromPy(val)

# create hdl bit value (for example STD_LOGIC value in vhdl)
hBit = lambda val: BIT.fromPy(val)

def mkRange(width):
    """Make hdl range (for example 1 downto 0 in vhdl)
       @return: (width -1, 0) 
    """
    to = toHVal(width)
    to = to - 1
    return to._downto(0)

def vecT(width, signed=None):
    """Make vector type with specified width for example
       std_logic_vector(width-1 downto 0) in vhdl
    """
    return Bits(widthConstr=mkRange(width), signed=signed, forceVector=True)

def vec(val, width, signed=None):
    """create hdl vector value"""
    assert val < evalParam(hInt(2)._pow(width)).val
    return vecT(width, signed).fromPy(val)



from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.eventCapableVal import EventCapableVal
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.value import Value, areValues
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc
from hwt.hdlObjects.types.bitVal_bitOpsVldMask import vldMaskForOr, \
    vldMaskForAnd
from operator import eq, ne

def boolLogOp__val(self, other, op, getVldFn):
    v = bool(op._evalFn(bool(self.val), (other.val)))
    return BooleanVal(v, BOOL,
            getVldFn(self, other),
            max(self.updateTime, other.updateTime))

def boolLogOp(self, other, op, getVldFn, whenOneIsVal):
    other = toHVal(other)

    if isinstance(self, Value):
        if isinstance(other, Value):
            return boolLogOp__val(self, other, op, getVldFn)

        return whenOneIsVal(self, other)
    else:
        if isinstance(other, Value):
            return whenOneIsVal(other, self)
        
    return Operator.withRes(op, [self, other._convert(BOOL)], BOOL)

def boolCmpOp__val(self, other, op, evalFn):
    v = evalFn(bool(self.val), bool(other.val)) and (self.vldMask == other.vldMask == 1)
    return BooleanVal(v, BOOL,
            self.vldMask & other.vldMask,
            max(self.updateTime, other.updateTime))
    
def boolCmpOp(self, other, op, evalFn=None):
    other = toHVal(other)
    if evalFn is None:
        evalFn = op._evalFn
    
    if areValues(self, other):
        return boolCmpOp__val(self, other, op, evalFn)
    else:
        return Operator.withRes(op, [self, other._convert(BOOL)], BOOL)

def whenOneIsVal_and(val, other):
    if val.vldMask:
        if val.val:
            return other
        else:
            v = val.clone()
            v.val = 0
            return v
    else:
        return val
    

def whenOneIsVal_or(val, other):
    if val.vldMask:
        if val.val:
            return val
        else:
            return other
    else:
        return val
    
class BooleanVal(EventCapableVal):
    
    @classmethod
    def fromPy(cls, val, typeObj):
        """
        @param val: value of python type bool or None
        @param typeObj: instance of HdlType
        """
        vld = int(val is not None)
        if not vld:
            val = False
        else:
            val = bool(val)
        return cls(val, typeObj, vld)

    def _eq__val(self, other):
        return boolCmpOp__val(self, other, AllOps.EQ, eq)
    def _eq(self, other):
        return boolCmpOp(self, other, AllOps.EQ, evalFn=eq)

    def _ne__val(self, other):
        return boolCmpOp__val(self, other, AllOps.NEQ, ne)
    def __ne__(self, other):
        return boolCmpOp(self, other, AllOps.NEQ)

    def _invert__val(self):
        v = self.clone()
        v.val = not v.val
        return v
    def __invert__(self):
        if isinstance(self, Value):
            return self._invert__val()
        else:
            try:
                # double negation
                d = self.singleDriver()
                if isinstance(d, Operator) and d.operator == AllOps.NOT:
                    return d.ops[0]
            except MultipleDriversExc:
                pass
            return Operator.withRes(AllOps.NOT, [self], BOOL)

    def _ternary__val(self, ifTrue, ifFalse):
        if self.val:
            if not self.vldMask:
                ifTrue.vldMask = 0
            return ifTrue
        else:
            if not self.vldMask:
                ifFalse.vldMask = 0
            return ifFalse
    def _ternary(self, ifTrue, ifFalse):
        ifTrue = toHVal(ifTrue)
        ifFalse = toHVal(ifFalse)
        
        if isinstance(self, Value):
            return self._ternary__val(ifTrue, ifFalse)
        else:
            return Operator.withRes(AllOps.TERNARY, [self, ifTrue, ifFalse], ifTrue._dtype)

    # logic
    def _and__val(self, other):
        return boolLogOp__val(self, other, AllOps.AND_LOG, vldMaskForAnd)
    def __and__(self, other):
        return boolLogOp(self, other, AllOps.AND_LOG, vldMaskForAnd, whenOneIsVal_and)

    def _or__val(self, other):
        return boolLogOp__val(self, other, AllOps.OR_LOG, vldMaskForOr)
    def __or__(self, other):
        return boolLogOp(self, other, AllOps.OR_LOG, vldMaskForOr, whenOneIsVal_or)


    # for evaluating only, not convertible to hdl
    def __bool__(self):
        assert isinstance(self, Value)
        return bool(self.val and self.vldMask)
    



from hwt.hdlObjects.hdlContext import HdlContext

class PackageHeader(HdlContext):
    """
    Hdl package container
    Entities means components from vhdl
    """
    def __init__(self, name, libraryCtx, isDummy=False):
        super(PackageHeader, self).__init__(name, libraryCtx)
        self.body = None
        self._isDummy = isDummy  
    
    def update(self, other):
        HdlContext.update(self, other)
        self.body = other.body
        
    def insertBody(self, body):
        self.body = body
        body.header = self
    
    def copyFrom(self, other):
        assert self._isDummy
        HdlContext.copyFrom(self, other)
        self._isDummy = False
        
class PackageBody(HdlContext):
    def __init__(self, name, libraryCtx):
        super(PackageBody, self).__init__(name, libraryCtx)
        self.header = None



import math
from operator import and_, or_
import types

from hwt.hdlObjects.constants import DIRECTION
from hwt.hdlObjects.operatorDefs import concatFn
from hwt.hdlObjects.typeShortcuts import hInt, vec, vecT
from hwt.hdlObjects.types.defs import BIT
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.pyUtils.arrayQuery import arr_any
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.signalUtils.walkers import discoverEventDependency
from hwt.synthesizer.vectorUtils import getWidthExpr, fitTo


def _intfToSig(obj):
    if isinstance(obj, InterfaceBase):
        return obj._sig
    else:
        return obj

class StmCntx(list):
    """
    Base class of statement contexts
    """
    pass


def flaten(iterables):
    if isinstance(iterables, (list, tuple, types.GeneratorType)):
        for i in iterables:
            yield from flaten(i)
    else:
        yield iterables


class If(StmCntx):
    """
    If statement generator
    
    @ivar nowIsEventDependent: flag if current scope of if is event dependent
    """
    def __init__(self, cond, *statements):
        """
        @param cond: condition in if
        @param statements: list of statements which should be active if condition is met   
        """
        self.cond = _intfToSig(cond)
        self.nowIsEventDependent = bool(list(discoverEventDependency(cond)))
        self.elifConds = []
        self._appendStatements(set([self.cond, ]), statements)
        
    def Else(self, *statements):
        ncond = set()
        ncond.add(~self.cond)
        
        for ec in self.elifConds:
            ncond.add(~ec)

        self._appendStatements(ncond, statements)
            
        # convert self to StmCntx to prevent any other else/elif
        stml = StmCntx()
        stml.extend(self)
        return stml
    
    def _appendStatements(self, condSet, statements):
        for stm in flaten(statements):
            stm.isEventDependent = stm.isEventDependent or self.nowIsEventDependent
            for c in condSet:
                c.endpoints.append(stm)
            stm.cond.update(condSet)
            self.append(stm)
    
    def Elif(self, cond, *statements):
        cond = _intfToSig(cond)
        self.nowIsEventDependent = self.nowIsEventDependent or\
                               arr_any(discoverEventDependency(cond), lambda x: True)
        thisCond = set()
        thisCond.add(~self.cond)
        for c in self.elifConds:
            thisCond.add(~c)
        thisCond.add(cond)
        
        self._appendStatements(thisCond, statements)
        
        self.elifConds.append(cond)
        
        return self


class Switch(StmCntx):
    """
    Switch statement generator
    """
    def __init__(self, switchOn):
        self.switchOn = switchOn
        self.cond = None
    
    _appendStatements = If._appendStatements
    
    def Case(self, caseVal, *statements):
        cond = self.switchOn._eq(caseVal)
        if self.cond is None:
            If.__init__(self, cond, *statements) 
        else:
            If.Elif(self, cond, *statements)
        return self
    
    def addCases(self, tupesValStmnts):
        s = self
        for val, statements in tupesValStmnts:
            if val is None:
                s = s.Default(*statements)
            else:
                s = s.Case(val, *statements)
        return s
    
    def Default(self, *statements):
        return If.Else(self, *statements)


def In(sigOrVal, iterable):
    """
    Hdl conversible in operator, check if any of items in "iterable" equals "sigOrVal"
    """
    res = None
    for i in iterable:
        i = toHVal(i)
        if res is None:
            res = sigOrVal._eq(i)
        else:
            res = res | sigOrVal._eq(i)

    return res


def _ForEach_callBody(fn, item, index):
    if fn.__code__.co_argcount == 1:
        return fn(item)
    else:
        return fn(item, index)


def ForEach(parentUnit, items, bodyFn, ack=None):
    """
    @param parentUnit: unit where this code should be instantiated
    @param items: items which this "for" itering on
    @param bodyFn: function which fn(item, index) or fn(item). 
                   It's content is performed in every iteration.
    @param ack: Ack signal to signalize that next iteration should be performed.
                if ack=None "for" will run for ever in loop 
    """
    
    items = list(items)
    l = len(items)
    if l == 0:
        # if there are no items there is nothing to generate
        return []
    elif l == 1:
        # if there is only one item do not generate counter logic generate
        return _ForEach_callBody(bodyFn, items[0], 0)
    else:
        # if there is multiple items we have to generate counter logic
        index = parentUnit._reg("for_index", vecT(log2ceil(l + 1), signed=False), defVal=0)
        incrIndex = index ** (index + 1)
        if ack is not None:
            If(ack,
               incrIndex
            )
        return Switch(index).addCases(
            [  (i, _ForEach_callBody(bodyFn, item, i)) 
                for i, item in enumerate(items)]
            ).Default(
                _ForEach_callBody(bodyFn, items[0], 0)
            )


class FsmBuilder(StmCntx):
    """
    @ivar stateReg: register with state
    """
    
    def __init__(self, parent, stateT, stateRegName="st"):
        """
        @param parent: parent unit where fsm should be builded
        @param stateT: enum type of state
        @param stateRegName: name of register where sate is stored
        """
        if isinstance(stateT, Enum):
            beginVal = stateT.fromPy(stateT._allValues[0])
        else:
            beginVal = 0
        
        self.stateReg = parent._reg(stateRegName, stateT, beginVal)
        Switch.__init__(self, self.stateReg)
    
    _appendStatements = Switch._appendStatements
    def Trans(self, stateFrom, *condAndNextState):
        """
        @param stateFrom: apply when FSM is in this state
        @param condAndNextState: tupes (condition, newState),
                        last does not to have condition
        
        @attention: transitions has priority, first has the biggest 
        @attention: if stateFrom is None it is evaluated as default
        """
        top = None
        last = True
        
        for cAndS in reversed(condAndNextState):
            if last is True:
                last = False
                # if this is last trans. it does not have to condition
                try:
                    condition, newvalue = cAndS
                except TypeError:
                    top = c(cAndS, self.stateReg)
                    continue
                top = [] 

            else:
                condition, newvalue = cAndS
            
            # building decision tree    
            top = If(condition,
                        c(newvalue, self.stateReg)
                    ).Else(
                        top
                    )
            
        if stateFrom is None:
            s = Switch.Default(self, *top)
        else:
            s = Switch.Case(self, stateFrom, *top)
        
        return s
    
    def Default(self, *condAndNextState):
        d = self.Trans(None, *condAndNextState)
        d.stateReg = self.stateReg
        return d


# class While(StmCntx):
#    def __init__(self, cond):
#        self.cnd = _intfToSig(cond)
#    
#    def Do(self, *statements):


def _connect(src, dst, exclude, fit):
        
    if isinstance(src, InterfaceBase):
        if isinstance(dst, InterfaceBase):
            return dst._connectTo(src, exclude=exclude, fit=fit)
        src = src._sig
        
    assert not exclude, "this intf. is just a signal"   
    if src is None:
        src = dst._dtype.fromPy(None)
    else:
        src = toHVal(src)
        
    if fit:
        src = fitTo(src, dst)
        
    src = src._dtype.convert(src, dst._dtype)
    
    return dst ** src


def connect(src, *destinations, exclude=set(), fit=False):
    """
    Connect src (signals/interfaces/values) to all destinations
    @param exclude: interfaces on any level on src or destinations 
                which should be excluded from connection process
    @param fit: auto fit source width to destination width 
    """
    assignemnts = []
    for dst in destinations:
        assignemnts.extend(_connect(src, dst, exclude, fit))
        
    return assignemnts


def packed(intf, masterDirEqTo=DIRECTION.OUT, exclude=set()):
    """
    Concatenate all signals to one big signal, recursively
    """
    if not intf._interfaces:
        if intf._masterDir == masterDirEqTo:
            return intf._sig
        return None
    
    res = None
    for i in intf._interfaces:
        if i in exclude:
            continue
        
        if i._interfaces:
            if i._masterDir == DIRECTION.IN:
                d = DIRECTION.opposite(masterDirEqTo)
            else:
                d = masterDirEqTo
            s = packed(i, d, exclude=exclude) 
        else:
            if i._masterDir == masterDirEqTo:
                s = i._sig
            else:
                s = None
        
        if s is not None:
            if res is None:
                res = s
            else:
                res = Concat(res, s)
        
    return res


def connectUnpacked(src, dst, exclude=[]):
    """src is packed and it is unpacked and connected to dst"""
    # [TODO] parametrized offsets
    offset = 0
    connections = []
    for i in reversed(list(walkPhysInterfaces(dst))):
        if i in exclude:
            continue
        sig = i._sig
        t = sig._dtype
        if t == BIT:
            s = src[hInt(offset)]
            offset += 1
        else:
            w = getWidthExpr(t)
            s = src[(w + offset): offset]
            offset += t.bit_length()
        connections.append(sig ** s)
    
    return connections


def packedWidth(intf):
    """Sum of all width of interfaces in this interface"""
    if isinstance(intf, type):
        # interface class
        intf = intf()
        intf._loadDeclarations()
    elif isinstance(intf, InterfaceBase) and not hasattr(intf, "_interfaces"):
        # not loaded interface
        _intf = intf

        intf = _intf.__class__()
        intf._updateParamsFrom(_intf)
        intf._loadDeclarations()
        
    
    if intf._interfaces:
        w = 0
        for i in intf._interfaces:
            w += packedWidth(i)
        return w
    else:
        t = intf._dtype
        if t == BIT:
            return 1
        return t.bit_length()


def _mkOp(fn): 
    def op(*ops):
        assert ops, ops
        top = None 
        for s in ops:
            if top is None:
                top = s
            else:
                top = fn(top, s)
        return top
    return op


# variadic operator functions
And = _mkOp(and_)
Or = _mkOp(or_)
Concat = _mkOp(concatFn)


def iterBits(sig):
    """
    Iterate over bits in vector
    """
    l = sig._dtype.bit_length() 
    for bit in range(l):
        yield sig[bit] 


def power(base, exp):
    return toHVal(base)._pow(exp) 


def ror(sig, howMany):
    """
    Rotate right
    """
    return  sig[howMany:]._concat(sig[:howMany])


def rol(sig, howMany):
    """
    Rotate left
    """
    l = sig._dtype.bit_length()
    b = l - howMany - 1
    return  sig[(l - howMany):]._concat(sig[:(l - howMany)])


def sll(sig, howMany):
    """
    Logical shift left
    """
    return vec(0, howMany)._concat(sig[:howMany])


def srl(sig, howMany):
    """
    Logical shift right
    """
    return sig[howMany:]._concat(vec(0, howMany))


def log2ceil(x):
    """
    Returns no of bits required to store x-1
    for example x=8 returns 3
    """
    
    if not isinstance(x, (int, float)):
        x = evalParam(x).val
    
    if x == 0 or x == 1:
        res = 1
    else:
        res = math.ceil(math.log2(x))
    return hInt(res)


def isPow2(num):
    assert isinstance(num, int)
    return num != 0 and ((num & (num - 1)) == 0)


def binToGray(sigOrVal):
    l = sigOrVal._dtype.bit_length()
    return Concat(sigOrVal[l - 1], sigOrVal[l - 1:0] ^ sigOrVal[l:1])


# shortcuts
c = connect



from hwt.serializer.constants import SERI_MODE
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import forAllParams
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import PropDeclrCollector
from hwt.synthesizer.interfaceLevel.unitImplHelpers import UnitImplHelpers
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist


class Unit(UnitBase, PropDeclrCollector, UnitImplHelpers):
    """
    Class members:
    #resolved automatically during configuration/declaration:

    @cvar _serializerMode: mode for serializer (drives when unit should be serialized)

    @ivar _interfaces: all interfaces 
    @ivar _units: all units defined on this obj in configuration/declaration
    @ivar _params: all params defined on this obj in configuration/declaration

    @ivar _checkIntferfaces: flag - after synthesis check if interfaces are present 
    @ivar _lazyLoaded : container of rtl object which were lazy loaded in implementation phase
                      (this object has to be returned from _toRtl of parent before it it's own objects)
    """

    _serializerMode = SERI_MODE.ALWAYS

    def __init__(self):
        self._checkIntferfaces = True
        self._lazyLoaded = []
        self._cntx = RtlNetlist(self)

        self._loadConfig()

    def _toRtl(self):
        """
        synthesize all subunits, make connections between them, build entity and component for this unit
        """
        assert not self._wasSynthetised()
        if not hasattr(self, "_name"):
            self._name = self._getDefaultName()
        self._cntx.globals = self._globalsFromParams()
        self._externInterf = []

        # prepare subunits
        for u in self._units:
            yield from u._toRtl()

        for u in self._units:
            subUnitName = u._name
            u._signalsForMyEntity(self._cntx, "sig_" + subUnitName)

        # prepare signals for interfaces
        for i in self._interfaces:
            signals = i._signalsForInterface(self._cntx)
            if i._isExtern:
                self._externInterf.extend(signals)

        self._loadMyImplementations()
        yield from self._lazyLoaded

        def forAllInterfaces(fn):
            for i in self._interfaces:
                fn(i)

            for u in self._units:
                for i in u._interfaces:
                    if i._isExtern:
                        fn(i)

        forAllInterfaces(lambda i : i._connectMyElems())

        if self._checkIntferfaces and not self._externInterf:
            raise Exception("Can not find any external interface for unit " + self._name\
                            + "- there is no such a thing as unit without interfaces")

        yield from self._synthetiseContext(self._externInterf)
        self._checkArchCompInstances()

    def _wasSynthetised(self):
        return self._cntx.synthesised

    def _synthetiseContext(self, externInterf):
        # synthesize signal level context
        s = self._cntx.synthesize(self._name, externInterf)
        self._entity = s[0]
        self._entity.__doc__ = self.__doc__
        self._entity.origin = self

        self._architecture = s[1]

        for intf in self._interfaces: 
            if intf._isExtern:
                intf._resolveDirections()
                # reverse because other components looks at this one from outside
                intf._reverseDirection()

        # connect results of synthesized context to interfaces of this unit
        self._boundInterfacesToEntity(self._interfaces)
        yield from s

        # after synthesis clean up interface so unit can be used elsewhere
        self._cleanAsSubunit()

    def _loadInterface(self, i, isExtern):
        i._loadDeclarations()
        i._setAsExtern(isExtern)

        if i._multipliedBy is not None:
            if i._interfaces:
                i._initArrayItems()
            else:
                i._injectMultiplerToDtype()

    def _loadDeclarations(self):
        """
        Load all declarations from _decl() method, recursively for all interfaces/units.
        """
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        if not hasattr(self, "_units"):
            self._units = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None
        for i in self._interfaces:
            self._loadInterface(i, True)

        # if I am a unit load subunits
        for u in self._units:
            u._loadDeclarations()

    def _registerIntfInImpl(self, iName, intf):
        """
        Register interface in implementation phase
        """
        self._registerInterface(iName, intf)
        self._loadInterface(intf, False)
        intf._signalsForInterface(self._cntx)

    @classmethod
    def _build(cls, multithread=True):
        """
        This Unit implementation does not require any preprocessing
        """
        pass

    def _globalsFromParams(self):
        # construct globals (generics for entity)
        globalNames = {}

        def addP(n, p):
            # [TODO] case sensitivity based on active HDL
            p.name = n.upper()
            n = n.lower()
            if n in globalNames:
                raise IntfLvlConfErr("Redefinition of generic '%s' while synthesis old:%s, new:%s" %
                                     (n, repr(globalNames[n]), repr(p)))
            globalNames[n] = p

        def nameForNestedParam(p):
            n = ""
            node = p
            while node is not self:
                if n == "":
                    n = node._name
                else:
                    n = node._name + "_" + n
                node = node._parent

            return n

        discoveredParams = set()
        for p in self._params:
            discoveredParams.add(p)
            addP(p.name, p)

        for intf in self._interfaces:
            for p in forAllParams(intf, discoveredParams):
                n = nameForNestedParam(p)
                addP(n, p)

        return globalNames

    def _getDefaultName(self):
        return self.__class__.__name__

    def _checkArchCompInstances(self):
        cInstances = len(self._architecture.componentInstances)
        units = len(self._units)
        if cInstances != units:
            inRtl = set(map(lambda x: x.name, self._architecture.componentInstances))
            inIntf = set(map(lambda x: x._name + "_inst", self._units))
            if cInstances > units:
                raise IntfLvlConfErr("_toRtl unit(s) %s were found in rtl but were not registered at %s" %
                                     (str(inRtl - inIntf), self._name))
            elif cInstances < units:
                raise IntfLvlConfErr("_toRtl of %s: unit(s) %s were lost" %
                                     (self._name, str(inIntf - inRtl)))



from hwt.simulator.utils import valueHasChanged
from hwt.hdlObjects.constants import DIRECTION, SENSITIVITY


def sensitivity(proc, *sensitiveTo):
    """
    register sensitivity for process
    """
    for s in sensitiveTo:
        if isinstance(s, tuple):
            sen, s = s
            if sen == SENSITIVITY.ANY:
                s.simSensProcs.add(proc)
            elif sen == SENSITIVITY.RISING:
                s.simRisingSensProcs.add(proc)
            elif sen == SENSITIVITY.FALLING:
                s.simRisingSensProcs.add(proc)
            else:
                raise AssertionError(sen)
        else:
            s.simSensProcs.add(proc)
         

def simEvalCond(simulator, *conds):
    """
    Evaluate list of values as condition
    """
    _cond = True
    _vld = True
    for v in conds:
        val = bool(v.val)
        fullVld = v.vldMask == 1
        if not val and fullVld:
            return False, True
        
        _cond = _cond and val
        _vld = _vld and fullVld
        
        
    return _cond, _vld

class SimModel(object):
    pass

def connectSimPort(simUnit, subSimUnit, srcName, dstName, direction):
    if direction == DIRECTION.OUT:
        origPort = getattr(subSimUnit, srcName)
        newPort = getattr(simUnit, dstName)
        setattr(subSimUnit, srcName, newPort)
    else:
        origPort = getattr(subSimUnit, dstName)
        newPort = getattr(simUnit, srcName)
        setattr(subSimUnit, dstName, newPort)
    
    subSimUnit._cntx.signals.remove(origPort)
    
    
def mkUpdater(nextVal, invalidate):
    """
    Create value updater for simulation
    """
    
    def updater(currentVal):
        _nextVal = nextVal.clone()
        if invalidate:
            _nextVal.vldMask = 0
        return (valueHasChanged(currentVal, _nextVal), _nextVal)
    return updater
            

def mkArrayUpdater(nextItemVal, indexes, invalidate):
    """
    Create value updater for simulation for value of array type
    """
    def updater(currentVal):
        if len(indexes) > 1:
            raise NotImplementedError()

        _nextItemVal = nextItemVal.clone()        
        if invalidate:
            _nextItemVal.vldMask = 0
            
        index = indexes[0]
        change = valueHasChanged(currentVal._getitem__val(index), _nextItemVal)
        currentVal._setitem__val(index, _nextItemVal) 
        return (change, currentVal)

    return updater



import imp
import importlib
import inspect
import os
import sys

from hwt.hdlObjects.constants import Time
from hwt.hdlObjects.types.defs import BIT
from hwt.serializer.simModel.serializer import SimModelSerializer
from hwt.simulator.agentConnector import autoAddAgents
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.simModel import SimModel
from hwt.simulator.simSignalProxy import IndexSimSignalProxy
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.shortcuts import toRtl, synthesised, toRtlAndSave


def simPrepare(unit, modelCls=None, dumpModelIn=None, onAfterToRtl=None):
    """
    Create simulation model and connect it with interfaces of original unit
    and decorate it with agents
    @return: tuple (fully loaded unit with connected sim model,
                    connected simulation model,
                    simulation processes of agents
                    )
    """
    if modelCls is None:
        modelCls = toSimModel(unit, tmpDir=dumpModelIn)
    else:
        synthesised(unit)

    if onAfterToRtl:
        onAfterToRtl(unit)

    reconectUnitSignalsToModel(unit, modelCls)
    model = modelCls()
    procs = autoAddAgents(unit)
    return unit, model, procs


def toSimModel(unit, tmpDir=None):
    """
    Create a simulation model for unit
    """
    if tmpDir is not None:
        toRtlAndSave(unit, tmpDir, serializer=SimModelSerializer)
        d = os.path.join(os.getcwd(), tmpDir)
        dInPath = d in sys.path
        if not dInPath:
            sys.path.append(d)

        simModule = importlib.import_module(unit._name)

        if not dInPath:
            sys.path.remove(d)
    else:
        sim_code = toRtl(unit, serializer=SimModelSerializer)
        simModule = imp.new_module('simModule')
        exec(sim_code, simModule.__dict__)

    return simModule.__dict__[unit._name]


def reconectUnitSignalsToModel(synthesisedUnitOrIntf, modelCls):
    """
    Reconnect model signals to unit to run simulation with simulation model
    but use original unit interfaces for communication
    """
    subInterfaces = synthesisedUnitOrIntf._interfaces
    reconnectArrayItems = isinstance(synthesisedUnitOrIntf, InterfaceBase) \
                            and synthesisedUnitOrIntf._multipliedBy is not None
    if subInterfaces:
        for intf in subInterfaces:
            reconectUnitSignalsToModel(intf, modelCls)
    else:
        s = synthesisedUnitOrIntf
        s._sigInside = getattr(modelCls, s._sigInside.name)

    if reconnectArrayItems:
        # if this interface is array we have to replace signals in array items as well
        for item in synthesisedUnitOrIntf._arrayElemCache:
            reconectArrayIntfSignalsToModel(synthesisedUnitOrIntf, item)


def reconectArrayIntfSignalsToModel(parent, item):
    index = parent._arrayElemCache.index(item)
    for p, i in zip(walkPhysInterfaces(parent), walkPhysInterfaces(item)):
        s = i._sigInside
        width = s._dtype.bit_length()
        if s._dtype == BIT:
            lowerIndex = None
            upperIndex = (width * index)
        else:
            lowerIndex = (width * index)
            upperIndex = (width * (index + 1))

        i._sigInside = IndexSimSignalProxy(i._name,
                                           p._sigInside,
                                           i._dtype,
                                           upperIndex,
                                           lowerIndex)


def simUnitVcd(simModel, stimulFunctions, outputFile=sys.stdout, time=100 * Time.ns):
    """
    Syntax sugar
    If outputFile is string try to open it as file
    @return: hdl simulator object
    """
    assert isinstance(simModel, SimModel), "Class of SimModel is required (got %r)" % (simModel)
    if isinstance(outputFile, str):
        d = os.path.dirname(outputFile)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFile, 'w') as f:
            return _simUnitVcd(simModel, stimulFunctions,
                               outputFile=f, time=time)
    else:
        return _simUnitVcd(simModel, stimulFunctions,
                           outputFile=outputFile, time=time)


def _simUnitVcd(simModel, stimulFunctions, outputFile, time):
    """
    @param unit: interface level unit to simulate
    @param stimulFunctions: iterable of function with single param env (simpy environment)
                            which are driving the simulation
    @param outputFile: file where vcd will be dumped
    @param time: endtime of simulation, time units are defined in HdlSimulator
    @return: hdl simulator object
    """
    sim = HdlSimulator()

    # configure simulator to log in vcd
    sim.config = VcdHdlSimConfig(outputFile)

    # run simulation, stimul processes are register after initial initialization
    sim.simUnit(simModel, time=time, extraProcesses=stimulFunctions)
    return sim


class CallbackLoop(object):
    def __init__(self, sig, condFn, fn):
        """
        @param sig: signal on which write callback should be used
        @param condFn: condition (function) which has to be satisfied in order to run callback
        @attention: if condFn is None callback function is always executed

        @ivar isGenerator: flag if callback function is generator or normal function
        """
        self.isGenerator = inspect.isgeneratorfunction(fn)
        self.condFn = condFn
        self.fn = fn
        try:
            # if sig is interface we need internal signal
            self.sig = sig._sigInside
        except AttributeError:
            self.sig = sig

    def onWriteCallback(self, sim):
        s = self.sig
        cond = self.condFn(s, sim)
        if cond is None or cond:
            if self.isGenerator:
                yield from self.fn(sim)
            else:
                self.fn(sim)
        s._writeCallbacks.append(self.onWriteCallback)
        # no function just asset this functionn will be generator
        yield sim.wait(0)

    def initProcess(self, sim):
        """
        Process for injecting of this callback loop into simulator
        """
        self.sig._writeCallbacks.append(self.onWriteCallback)
        yield sim.wait(0)


def isRising(sig, sim):
    return bool(sim.read(sig)._onRisingEdge(sim.now))


def onRisingEdge(sig, fn):
    """
    Call function (or generator) everytime when signal is on rising edge
    """
    c = CallbackLoop(sig, isRising, fn)
    return c.initProcess


def onRisingEdgeNoReset(sig, reset, fn):
    """
    Call function (or generator) everytime when signal is on rising edge and reset
    is not active
    """
    def cond(clk, sim):
        r = sim.read(reset)
        if reset.negated:
            r.val = not r.val
        return isRising(clk, sim) and not bool(r)

    c = CallbackLoop(sig, isRising, fn)
    return c.initProcess


def oscilate(sig, period=10 * Time.ns, initWait=0):
    """
    Oscilative simulation driver for your signal
    """
    halfPeriod = period / 2

    def oscilateStimul(s):
        s.write(False, sig)
        yield s.wait(initWait)

        while True:
            yield s.wait(halfPeriod)
            s.write(True, sig)
            yield s.wait(halfPeriod)
            s.write(False, sig)
    return oscilateStimul


def pullDownAfter(sig, intDelay=6 * Time.ns):
    """
    @return: simulation driver which keeps value high for intDelay then it sets
             value to 0
    """
    def _pullDownAfter(s):
        s.write(True, sig)
        yield s.wait(intDelay)
        s.write(False, sig)

    return _pullDownAfter


def pullUpAfter(sig, intDelay=6 * Time.ns):
    """
    @return: Simulation driver which keeps value low for intDelay then it sets
             value to 1
    """
    def _pullDownAfter(s):
        s.write(False, sig)
        yield s.wait(intDelay)
        s.write(True, sig)

    return _pullDownAfter



import os, fnmatch

def find_files(directory, pattern, recursive=True):
    if not os.path.isdir(directory):
        if os.path.exists(directory):
            raise IOError(directory + ' is not directory')
        else:
            raise IOError(directory + " does not exists")
    if recursive:
        for root, _, files in os.walk(directory):
            for basename in files:
                if fnmatch.fnmatch(basename, pattern):
                    filename = os.path.join(root, basename)
                    yield filename
    else:
        root = directory
        for basename in os.listdir(root):
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename
        
            
def applyReplacesOnFile(fileName, replaces, write=True):
    with open(fileName) as f:
        content = f.read()
    
    for r in replaces:
        content = content.replace(r, replaces[r])    
    if write:    
        with open(fileName, "w") as f:
            f.write(content)
    else:
        return content

class ChDir:
    """cd with backtrack"""         
    def __init__(self, newPath):  
        self.newPath = newPath
    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, type, value, tb):
        os.chdir(self.savedPath)


from copy import copy

from hwt.hdlObjects.constants import Unconstrained
from hwt.hdlObjects.types.hdlType import HdlType 
from hwt.bitmask import mask


class Bits(HdlType):
    
    def __init__(self, widthConstr=None, forceVector=False, signed=None):
        """
        @param forceVector: use always hdl vector type 
            (for example std_logic_vector(0 downto 0) instead of std_logic)
        """
        super().__init__()
        self.forceVector = forceVector
        self.signed = signed
        if widthConstr is None:
            self.constrain = Unconstrained()
        else:
            self.constrain = widthConstr
    
    def __eq__(self, other):
        return isinstance(other, Bits) and other.bit_length() == self.bit_length()\
            and self.signed == other.signed and self.forceVector == other.forceVector
    
    def __hash__(self):
        return hash((self.signed, id(self.constrain), self.forceVector))
    
    def applySpecificator(self, const):
        assert isinstance(self.constrain, Unconstrained)
        s = copy(self)
        s.constrain = const
        return s
    
    def all_mask(self):
        return mask(self.bit_length())
    
    def bit_length(self):
        if isinstance(self.constrain, (int, float)):
            return int(self.constrain)
        elif isinstance(self.constrain, Unconstrained):
            try:
                return self.constrain.derivedWidth
            except AttributeError:
                return None
        else:
            w = self.constrain.staticEval()
            return abs(w.val[0].val - w.val[1].val) + 1 

    
    
    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.bitsConversions import convertBits
        return convertBits
    
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.bitsVal import BitsVal 
            cls._valCls = BitsVal
            return cls._valCls

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        c = self.constrain
        if isinstance(c, int):
            constr = "width:%d" % c
        elif isinstance(c, Unconstrained):
            try:
                constr = "derivedWidth:%d" % (c.derivedWidth)
            except AttributeError:
                constr = ""
        else:
            constr = VhdlSerializer.asHdl(self.constrain, onlyPrintDefaultValues)
        
        return "<HdlType %s, %s>" % (
            self.__class__.__name__, constr)




from hwt.hdlObjects.value import Value, areValues
from hwt.hdlObjects.types.defs import BOOL, INT, SLICE
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.typeCast import toHVal
from hwt.hdlObjects.types.integer import Integer
from operator import pow, add, sub, mul, floordiv, eq, ne, le, lt, ge, gt

BoolVal = BOOL.getValueCls()
SliceVal = SLICE.getValueCls()

def intOp__val(self, other, op, resT, evalFn):
    v = evalFn(self.val, other.val)
    vldMask = int(self.vldMask and other.vldMask)
    updateTime = max(self.updateTime, other.updateTime)
    return resT.getValueCls()(v, resT, vldMask, updateTime)

def intOp(self, other, op, resT, evalFn=None):
    if evalFn is None:
        evalFn = op._evalFn
        
    other = toHVal(other)._convert(INT)
    if areValues(self, other):
        return intOp__val(self, other, op, resT, evalFn)
    else:
        return Operator.withRes(op, [self, other], resT)

def intAritmeticOp(self, other, op):
    return intOp(self, other, op, INT)

def intCmpOp(self, other, op, evalFn=None):
    return intOp(self, other, op, BOOL, evalFn=evalFn)

class IntegerVal(Value):
    """
    @ivar vldMask: can be only 0 or 1
    @ivar updateTime: time when this value was set, used in simulator
    """
    @classmethod
    def fromPy(cls, val, typeObj):
        """
        @param val: value of python type int or None
        @param typeObj: instance of HdlType
        """
        assert isinstance(typeObj, Integer)
        vld = int(val is not None)
        if not vld:
            val = 0
        else:
            val = int(val)
        
        return cls(val, typeObj, vld)
    
    # used only as syntax shugar for simulations
    def __int__(self):
        if self.vldMask:
            return self.val
        else:
            return None
    
    # arithmetic
    def _neg__val(self):
        v = self.clone()
        v.val = -self.val 
        return v
    
    def __neg__(self):
        if isinstance(self, Value):
            return self._neg__val()
        else:
            return Operator.withRes(AllOps.UN_MINUS, [self], INT)

    def _add__val(self, other):
        return intOp__val(self, other, AllOps.ADD, INT, add)
    def __add__(self, other):
        return intAritmeticOp(self, other, AllOps.ADD)
        
    def _sub__val(self, other):
        return intOp__val(self, other, AllOps.SUB, INT, sub)
    def __sub__(self, other):
        return intAritmeticOp(self, other, AllOps.SUB)
    
    def _mul__val(self, other):
        return intOp__val(self, other, AllOps.MUL, INT, mul)
    def __mul__(self, other):
        return intAritmeticOp(self, other, AllOps.MUL)

    def _pow__val(self, other):
        return intOp__val(self, other, AllOps.POW, INT, pow)
    def _pow(self, other):
        return intOp(self, other, AllOps.POW, INT, pow)
    
    def _floordiv__val(self, other):
        return intOp__val(self, other, AllOps.DIV, INT, floordiv)
    def __floordiv__(self, other):
        return intAritmeticOp(self, other, AllOps.DIV)
    
    
    def _downto__val(self, other):
        vldMask = int(self.vldMask and other.vldMask)
        updateTime = max(self.updateTime, other.updateTime)
        return SliceVal((self, other), SLICE, vldMask, updateTime)
    def _downto(self, other):
        other = toHVal(other)._convert(INT)
        if areValues(self, other):
            return self._downto__val(other)
        else:
            return Operator.withRes(AllOps.DOWNTO, [self, other], SLICE)
    
    # comparisons    
    def _eq__val(self, other):
        return intOp__val(self, other, AllOps.EQ, BOOL, eq)
    def _eq(self, other):
        return intCmpOp(self, other, AllOps.EQ, eq)
    
    def _ne__val(self, other):
        return intOp__val(self, other, AllOps.NEQ, BOOL, ne)
    def __ne__(self, other):
        return intCmpOp(self, other, AllOps.NEQ)

    def _le__val(self, other):
        return intOp__val(self, other, AllOps.LE, BOOL, le)
    def __le__(self, other):
        return intCmpOp(self, other, AllOps.LE)
    
    def _lt__val(self, other):
        return intOp__val(self, other, AllOps.LOWERTHAN, BOOL, lt)
    def __lt__(self, other):
        return intCmpOp(self, other, AllOps.LOWERTHAN)
    
    def _ge__val(self, other):
        return intOp__val(self, other, AllOps.GE, BOOL, ge)
    def __ge__(self, other):
        return intCmpOp(self, other, AllOps.GE)
    
    def _gt__val(self, other):
        return intOp__val(self, other, AllOps.GREATERTHAN, BOOL, gt)



from hwt.hdlObjects.types.hdlType import HdlType 

class Enum(HdlType):
    def __init__(self, name, valueNames):
        super(Enum, self).__init__()
        self.name = name
        self._allValues = valueNames
        for n in valueNames:
            v = self.fromPy(n)
            setattr(self, n, v)
    
    def _add(self, name):
        """
        Add member of the enum
        """
        self._allValues.append(name)
        v = self.fromPy(name)
        setattr(self, name, v)
        return v
    
    def all_mask(self):
        return 1
    
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.enumVal import EnumVal 
            cls._valCls = EnumVal
            return cls._valCls